//
//  AppDelegate.swift
//  iOS Navigate Reference App
//
//  Created by Jonda, Sachin on 04.12.22.
//  Copyright © 2023 HERE Technologies. All rights reserved.
//

import UIKit
import heresdk

@UIApplicationMain
class AppDelegate: UIResponder, UIApplicationDelegate {

    func application(_ application: UIApplication, didFinishLaunchingWithOptions launchOptions: [UIApplication.LaunchOptionsKey: Any]?) -> Bool {
        // We sugest calling this explicitly.
        // Initialization is a synchronous call, so if no error is thrown, SDK is guaranteed to be initialized.
        // When using mapview, you can also skip this, since mapview would also init the SDK for cenvenience reasons.

        // prevent screen lock for this application
        application.isIdleTimerDisabled = true

        // central visual configuration if ui segmented controller (button switch) to custom colors
        UISegmentedControl.appearance().selectedSegmentTintColor = .lightGray
        UISegmentedControl.appearance().setTitleTextAttributes([.foregroundColor: UIColor.white], for: .selected)
        UISegmentedControl.appearance().setTitleTextAttributes([.foregroundColor: UIColor.gray], for: .normal)

        // Initialize manually with options to set filesystem cache path with default system cache path.
        // Happens also indirectly when using map view and credentials are set in plist, but I do this here more explicity for demonstration purposes
        let cachePath = FileManager.default.urls(for: .cachesDirectory, in: .userDomainMask).first?.path ?? FileManager.default.temporaryDirectory.path

        // reading credentials from info.plist in this usecase
        let credits = Bundle.main.object(forInfoDictionaryKey: "HERECredentials") as? Dictionary<String,String>
        guard credits != nil else {
            print("ERROR: No HERE credentials set in info.plist")
            return false
        }
        let key = credits?["AccessKeyId"]
        let secret = credits?["AccessKeySecret"]
        guard key != nil && secret != nil else {
            print("ERROR: No HERE credentials set in info.plist")
            return false
        }

        print("Initializing SDK with credentials \(String(describing: key))")

        var options = SDKOptions(   accessKeyId: key!,
                                    accessKeySecret: secret!,
                                    cachePath: cachePath,
                                    cacheSizeInBytes: 1024*1024*256) // 256 MB temporary cache size for streamed data

        options.politicalView = "USA"  // setting political view explicitly to international, like USA

        /*
        options.layerConfiguration = LayerConfiguration(enabledFeatures: [LayerConfiguration.Feature.detailRendering,
                                                                          LayerConfiguration.Feature.navigation,
                                                                          LayerConfiguration.Feature.offlineSearch,
                                                                          LayerConfiguration.Feature.offlineRouting,
                                                                          LayerConfiguration.Feature.rendering,
                                                                          LayerConfiguration.Feature.truck])
         */

        // Automatically keep map cache up to date when new map versions are available;
        // This will vanish automatically cached map when update appears
        options.autoUpdateOfOnlineCache = false
        // TODO: Use https://main.gitlab.in.here.com/here-sdk/sdk/-/merge_requests/550/diffs?commit_id=23d8809999306e65f1ddee5315a75ad095d71dfd instead ?

        do {
            try SDKNativeEngine.makeSharedInstance(options: options)
        } catch {
            fatalError("Failed to initialize SDK. Cause: \(error)")
        }

        return true
    }

    // MARK: UISceneSession Lifecycle

    func application(_ application: UIApplication, configurationForConnecting connectingSceneSession: UISceneSession, options: UIScene.ConnectionOptions) -> UISceneConfiguration {
        // Called when a new scene session is being created.
        // Use this method to select a configuration to create the new scene with.
        return UISceneConfiguration(name: "Default Configuration", sessionRole: connectingSceneSession.role)
    }

    func application(_ application: UIApplication, didDiscardSceneSessions sceneSessions: Set<UISceneSession>) {
        // Called when the user discards a scene session.
        // If any sessions were discarded while the application was not running, this will be called shortly after application:didFinishLaunchingWithOptions.
        // Use this method to release any resources that were specific to the discarded scenes, as they will not return.
    }

    func applicationWillTerminate(_ application: UIApplication) {
            // we should cleanup the MapView when we kill the app
            print("App gets terminated...Cleaning up MapView")
            MapView.deinitialize()
            print("MapView deinitialization done.")
    }

    func applicationWillResignActive(_ application: UIApplication) {
        print("App enters background...Stopping MapView rendering")
        MapView.pause()
    }

    func applicationDidBecomeActive(_ application: UIApplication) {
        print("App enters foreground...Resuming MapView rendering")
        MapView.resume()
    }

}

