//
//  ContentView.swift
//  iOS Navigate Reference App
//
//  Created by Jonda, Sachin on 04.12.22.
//  Copyright © 2023 HERE Technologies. All rights reserved.
//

import SwiftUI
import heresdk
import SVGView

struct ContentView: View {

    private var hereMap = HereMapView()

    @ObservedObject private var mapSettings: MapSettings
    @ObservedObject private var mapController: MapController

    @State private var isMenuOpen = false

    init(settings: MapSettings){
        self.mapSettings = settings
        self.mapController = MapController(settings: settings)
    }
    
    var body: some View {
        ZStack (alignment: .center){

            GeometryReader { geometry in

                ZStack (alignment: .leading){

                    // HERE Map
                    self.hereMap
                        .onAppear {
                            self.mapController.setupMap(mapView: self.hereMap.mapViewHolder.mapView, mapSettings: self.mapSettings)
                        }
                        .onDisappear(){
                            print("MapView disappears...")
                        }
                        .offset(x: self.isMenuOpen ? geometry.size.width/1.5 : 0)
                        .disabled(self.isMenuOpen ? true : false)  // disable map interaction when menu is open
                        .edgesIgnoringSafeArea(.all)

                    // Map overlay UI components (search, guidance bar, buttons, etc.)
                    VStack {

                        // top bar
                        if self.mapController.isNavigationInProgress && self.mapController.isRouteAvailable {
                            NavigationDashboard(mapController: self.mapController)
                                .cornerRadius(10, corners: [.allCorners])
                                .shadow(color: .gray, radius: 2, x: 3, y: 3)
                                .padding(.vertical)
                        } else if self.mapController.isNavigationInProgress == false {
                            SearchBar(mapController: self.mapController, menuOpen: self.$isMenuOpen)
                                .background(Color(#colorLiteral(red: 0, green: 0.3, blue: 0, alpha: 1)))
                                .cornerRadius(10, corners: [.allCorners])
                                .shadow(color: .gray, radius: 2, x: 3, y: 3)
                                .padding(.vertical)
                        }

                        // show map scheme selector only when in explore mode - not in guidance (we need the space for alne assistance)
                        if self.mapController.isNavigationInProgress == false {
                            HStack{
                                Spacer()
                                
                                // mapscheme switch
                                Button(action: {
                                    print("Changing mapscheme...")
                                    self.mapSettings.showingSchemeSwitcher.toggle()
                                } )
                                {
                                    Image("day_normal")
                                        .resizable()
                                        .clipShape(Circle())
                                        .shadow(color: .gray, radius: 2, x: 3, y: 3)
                                        .overlay(Circle().stroke(Color.white, lineWidth: 3))
                                }
                                .frame(width: 50)
                            }
                            .frame(height: 50)
                        }
                        
                        if self.mapController.nextLaneInformation.count <= 0 {
                            LaneInformationView(nextLaneInformation: self.mapController.nextLaneInformation)
                                .frame(height: 50)
                        }
                               
                        // RealisticView over map
                        ZStack {  // JV and signpost are intended to be stacked o top of each other.
                            if self.mapController.junctionView != nil {
                                SVGView(string: self.mapController.junctionView!)
                                    .aspectRatio(contentMode: .fit)
                            }
                            
                            if self.mapController.signPost != nil {
                                SVGView(string: self.mapController.signPost!)
                                    .aspectRatio(contentMode: .fit)
                            }
                            
                        }
                        .frame(width: UIScreen.main.bounds.width*0.8)
                        .frame(height: 200)
                        //.border(Color.white)
                        
                        Spacer()

                        // lower button bar
                        HStack {

                            // compass & mypos button
                            VStack {
                                // compass button
                                Button(action: {
                                    self.mapController.northUp()
                                } )
                                {
                                    Image(systemName:"safari.fill")
                                        .resizable()
                                        .foregroundColor(Color(#colorLiteral(red: 0, green: 0, blue: 0, alpha: 1)))
                                        .background(Color(#colorLiteral(red: 1, green: 1, blue: 1, alpha: 1)))
                                        .frame(width: 30.0,height: 30.0)
                                        .clipShape(Circle())
                                        .shadow(color: .gray, radius: 2, x: 3, y: 3)
                                        .overlay(Circle().stroke(Color.white, lineWidth: 3))
                                }
                                .padding(.bottom, 5)

                                // my position button
                                Button(action: {
                                    self.mapController.zoomToMyPosition()
                                } )
                                {
                                   Image(systemName:"mappin.circle.fill")
                                        .resizable()
                                        .foregroundColor(Color(#colorLiteral(red: 0, green: 0, blue: 0, alpha: 1)))
                                        .background(Color(#colorLiteral(red: 1, green: 1, blue: 1, alpha: 1)))
                                        .frame(width: 30.0, height: 30.0)
                                        .clipShape(Circle())
                                        .shadow(color: .gray, radius: 2, x: 3, y: 3)
                                        .overlay(Circle().stroke(Color.white, lineWidth: 3))
                                }

                            }

                            Spacer()

                            // attention & warning dialog. mainly used for safety cameras.
                            if self.mapController.attentionDialog != nil {

                                VStack(alignment: .leading) {
                                           Text(self.mapController.attentionDialog!)
                                            .font(.headline)
                                            .foregroundColor(Color(#colorLiteral(red: 1, green: 1, blue: 1, alpha: 1)))
                                            .padding(.horizontal)
                                            .lineLimit(2)
                                }
                                .frame(minWidth: 0, maxWidth: .infinity, minHeight: 0, maxHeight: 60, alignment: .center )
                                .background(Color(#colorLiteral(red: 0, green: 0.3, blue: 0, alpha: 1)))
                                .cornerRadius(10, corners: [.allCorners])
                                .shadow(color: .gray, radius: 2, x: 3, y: 3)
                                .padding(.horizontal)

                                Spacer()
                            }

                            // speed limit sign
                            if self.mapController.isNavigationInProgress {
                                SpeedLimitSign( speedLimitInMetersPerSecond: self.mapController.currentSpeedLimit)
                                    .scaleEffect(self.mapController.isSpeeding ? 1.1 : 1.0) // if speeding, make it bigger with bouncy animation to catch attention
                                    .frame(width: 70)
                            }
                        }
                        .frame(height: 70)
                        .padding(.bottom)

                        // bottom navigation bar
                        NavigationStatusBar(mapController: self.mapController, mapSettings: self.mapSettings)
                            .background( self.mapController.isNavigationInProgress ? SwiftUI.Color(#colorLiteral(red: 0, green: 0.3, blue: 0, alpha: 1)) : Color.clear )
                            .cornerRadius(10, corners: [.allCorners])
                            .shadow(color: .gray, radius: 2, x: 3, y: 3)
                            .padding(.bottom)

                    }
                    .offset(x: self.isMenuOpen ? geometry.size.width/1.5 : 0)
                    .padding(.horizontal)

                    // side menu for options
                    if self.isMenuOpen {
                        DrawerMenu(mapSettings: self.mapSettings)
                        .frame(width: geometry.size.width/1.5)
                        .padding(.top, geometry.safeAreaInsets.top)
                        .padding(.bottom, geometry.safeAreaInsets.bottom)
                        .transition(.move(edge: .leading))
                        .edgesIgnoringSafeArea(.all)
                        .gesture( DragGesture()
                        .onEnded {
                            if $0.translation.width < -50 {  // if we swip left on menu for more than 50sceenpixels length, close menu again
                                withAnimation {
                                    self.isMenuOpen = false
                                }
                            }
                        })
                    }

                    // overlays
                    HStack {
                        Spacer()

                        // mapscheme / mapfeature switcher
                        if self.mapSettings.showingSchemeSwitcher {
                            MapSchemeSelectorView(mapController: self.mapController, mapSettings: self.mapSettings)
                                .frame(width: geometry.size.width/1.2, height: geometry.size.height/2.5)
                        }

                        // modal progress dialog !
                        if self.mapController.isWaitingDialogShown {
                            WaitingDialog(titleText: self.mapController.waitingDialogText)
                                .frame(width: geometry.size.width/1.5, height: geometry.size.height/3)
                        }

                        Spacer()
                    }

                } // inner ZStack

                // show map loader view - full screen - treated as seperate screen
                if self.mapSettings.showingMapDownloader {
                    MapLoaderView()
                        .environmentObject( self.mapController.mapLoaderController! )
                }

                // show map updater view - full screen - treated as seperate screen
                if self.mapSettings.showingMapUpdater {
                    MapUpdaterView()
                        .environmentObject( self.mapController.mapUpdaterController! )
                }

            }  // GeometryReader

        }  // outter ZStack
        .alert(isPresented: self.$mapController.isAlertShown, content: {  // alert dialog if something is wrong and user needs to get notified
            Alert(title: Text("Warning"), message: Text(self.mapController.alertMessage), dismissButton: .default(Text("OK")))
        })

    }
    
}


struct ContentView_Previews: PreviewProvider {
    static var previews: some View {
        ContentView(settings: MapSettings())
    }
}
