//
//  DrawerMenuView.swift
//
//  Created by Jonda, Sachin on 10.06.22.
//  Copyright © 2023 HERE Technologies. All rights reserved.
//

import SwiftUI
import heresdk

struct DrawerMenu: View {

    @ObservedObject var mapSettings: MapSettings
    
    @State var showPicker = false
    @State var trashStateColor = SwiftUI.Color.white

    var body: some View {
        GeometryReader { geometry in
            NavigationView {
                 ScrollView( .vertical, showsIndicators: true ) {
                        VStack(alignment: .leading, spacing: 0) {

                            // Header of side menu
                            VStack (alignment: .leading){
                                Image("HERE Logo white")
                                    .resizable()
                                    .frame(width: 100.0, height: 90.0)
                                    .padding(.bottom)

                                Text("HERE SDK Navigate")
                                    .font(.subheadline)
                                    .foregroundColor(Color.white)
                                Text("iOS REFERENCE")
                                    .font(.title)
                                    .foregroundColor(Color.white)
                                Text("APP")
                                    .font(.subheadline)
                                    .foregroundColor(Color.white)
                            }
                            .padding(.bottom)

                            // Transport Mode picker
                            VStack (alignment: .leading){

                                if self.mapSettings.transportMode==TransportMode.car {
                                    Text("Transportmode: Car")
                                    .foregroundColor(.white)
                                } else if self.mapSettings.transportMode==TransportMode.truck {
                                    Text("Transportmode: Truck")
                                    .foregroundColor(.white)
                                } else if self.mapSettings.transportMode==TransportMode.scooter {
                                    Text("Transportmode: Scooter")
                                    .foregroundColor(.white)
                                } else if self.mapSettings.transportMode==TransportMode.bicycle {
                                    Text("Transportmode: Bicycle")
                                    .foregroundColor(.white)
                                } else {
                                    Text("Transportmode: Pedestrian")
                                    .foregroundColor(.white)
                                }

                                Picker(selection: self.$mapSettings.transportMode, label: EmptyView()) {
                                    Image(systemName: "car")
                                        .imageScale(.medium)
                                        .tag(TransportMode.car)

                                    Image("TransportModePanel.truck")
                                        .imageScale(.medium)
                                        .tag(TransportMode.truck)

                                    Image("TransportModePanel.scooter")
                                        .imageScale(.medium)
                                        .tag(TransportMode.scooter)

                                    Image(systemName: "bicycle")
                                        .imageScale(.medium)
                                        .tag(TransportMode.bicycle)

                                    Image(systemName: "figure.walk")
                                        .imageScale(.medium)
                                        .tag(TransportMode.pedestrian)
                                }
                                .pickerStyle(SegmentedPickerStyle())
                            }
                            .padding([.top, .bottom])

                            // Other option switches
                            VStack (alignment: .leading){

                                Toggle(isOn: self.$mapSettings.simulationMode) {
                                    Text("Simulate navigation:")
                                        .foregroundColor(.white)
                                        .font(.subheadline)
                                }

                                Toggle(isOn: self.$mapSettings.recordingActive) {
                                    Text("Record GPX:")
                                        .foregroundColor(.white)
                                        .font(.subheadline)
                                }

                            }
                            .padding([.top, .bottom, .trailing])  // add extra trailing padding due to ugly artefacts when switch is too close to the border

                            // Offline options
                            VStack (alignment: .leading){

                                Toggle(isOn: self.$mapSettings.offlineMode) {
                                    Text("SDK full offline mode:")
                                        .foregroundColor(.white)
                                        .font(.subheadline)
                                }
                                
                                Toggle(isOn: self.$mapSettings.useOfflineSearch) {
                                    Text("   - Search offline:")
                                        .foregroundColor(.white)
                                        .font(.subheadline)
                                }

                                Toggle(isOn: self.$mapSettings.useOfflineRouting) {
                                    Text("   - Route offline:")
                                        .foregroundColor(.white)
                                        .font(.subheadline)
                                }

                                HStack {
                                        Text("Preload maps")
                                            .foregroundColor(Color.white)
                                            .font(.subheadline)

                                        Spacer()

                                        Image(systemName: "arrow.down.square")
                                            .imageScale(.large)
                                            .foregroundColor(.white)
                                }
                                .onTapGesture {
                                    print("Clicked MapDownloader...")
                                    self.mapSettings.showingMapDownloader = true
                                }
                                .padding(.top, 15)

                                HStack {
                                        Text("Check for map updates")
                                            .foregroundColor(Color.white)
                                            .font(.subheadline)

                                        Spacer()

                                        Image(systemName: "goforward")
                                            .imageScale(.large)
                                            .foregroundColor(.white)
                                }
                                .onTapGesture {
                                    print("Clicked MapUpdater...")
                                    self.mapSettings.showingMapUpdater = true
                                }
                                .padding(.top, 5)

                                HStack {
                                        Text("Delete cache")
                                            .foregroundColor(Color.white)
                                            .font(.subheadline)

                                        Spacer()

                                        Image(systemName: "trash")
                                            .imageScale(.large)
                                            .foregroundColor(self.trashStateColor)
                                }
                                .padding(.top, 5)
                                .simultaneousGesture(
                                    DragGesture(minimumDistance: 0)
                                        .onChanged({ _ in
                                            // pressed button
                                            self.trashStateColor = Color.red
                                        })
                                        .onEnded({ _ in
                                            // released button
                                            print("Clicked wiping temporary cache...")
                                            self.trashStateColor = .white
                                            let impactHeavy = UIImpactFeedbackGenerator(style: .heavy)
                                            impactHeavy.impactOccurred()
                                            self.mapSettings.clearCache()
                                        })
                                )

                            }
                            .padding([.top, .bottom, .trailing])  // add extra trailing padding due to ugly artefacts when switch is too close to the border

                            Spacer()

                            // Debug information like language and app version
                            VStack(alignment: .leading) {
                                Text("Debug information:")
                                    .fontWeight(.bold)
                                    .underline()
                                    .foregroundColor(Color.white)
                                    .padding(.bottom)

                                // debug info list. swiftui has a limit of 10 view items in a stack; so i have to group it!
                                Group {
                                    Text("App language: \(self.mapSettings.localeAppLanguage?.description ?? "unknown")")
                                        .font(.footnote)
                                        .foregroundColor(Color.white)

                                    Text("Device language: \(self.mapSettings.localeDeviceLanguage.description)")
                                        .font(.footnote)
                                        .foregroundColor(Color.white)

                                    Text("Device region: \(self.mapSettings.localeDeviceRegion?.description ?? "unknown")")
                                        .font(.footnote)
                                        .foregroundColor(Color.white)

                                    Text("Map language: \(self.mapSettings.localeDeviceLanguageMapped?.name ?? "unsupported")")
                                        .font(.footnote)
                                        .foregroundColor(Color.white)

                                    Text("App version: \(self.mapSettings.appVersion ?? "n/a") / build: \(self.mapSettings.appBuildNumber ?? "n/a")")
                                        .font(.footnote)
                                        .foregroundColor(Color.white)

                                    Text("SDK version: \(SDKBuildInformation.sdkVersion().versionName)")
                                        .font(.footnote)
                                        .foregroundColor(Color.white)

                                    Text("SDK build number: \(SDKBuildInformation.sdkVersion().versionBuild)")
                                        .font(.footnote)
                                        .foregroundColor(Color.white)

                                    Text("SDK build flavor: \(SDKBuildInformation.sdkVersion().versionTag)")
                                        .font(.footnote)
                                        .foregroundColor(Color.white)

                                    Text("SDK variant: \(SDKBuildInformation.sdkVersion().productVariant)")
                                        .font(.footnote)
                                        .foregroundColor(Color.white)
                                }
                            }
                            .padding([.top, .bottom])
                        }
                        .padding([.leading, .trailing])
                 }
                 .padding(.bottom, geometry.safeAreaInsets.bottom )
                 .padding(.top, geometry.safeAreaInsets.top )
                 .background(Color(#colorLiteral(red: 0.2549019754, green: 0.2745098174, blue: 0.3019607961, alpha: 1)))
                 .edgesIgnoringSafeArea(.all)
            }
        }
    }
}

struct DrawerMenuView_Previews: PreviewProvider {
    static var previews: some View {
        DrawerMenu(mapSettings: MapSettings())
    }
}
