//
//  HERELocationProvider.swift
//  iOS Navigate Reference App
//
//  Created by Jonda, Sachin on 24.03.22.
//  Copyright © 2023 HERE Technologies. All rights reserved.
//

import Foundation
import CoreLocation
import SwiftUI
import heresdk

class HERELocationProvider :    NSObject,
                                LocationStatusDelegate,
                                LocationDelegate,
                                CLLocationManagerDelegate
{

    var engine: LocationEngine?
    private var platformLocationService = CLLocationManager()  // used for authorization only

    var navigationDelegate: LocationDelegate?
    var appDelegate: LocationDelegate?


    // Comment:
    // On iOS, this is same as https://developer.apple.com/documentation/corelocation/cllocationmanager/1423836-desiredaccuracy
    // We use here by default 'navigation' but this uses all sensors all the time. It's best positioning, but most battery consuming, even if in map mode only
    // In practice, we switch this more flexible, so use reduced or less battery gungry accuracy in map, and switch to navigate only on demand / in navigatio mode
    private var locationAccuracy = LocationAccuracy.navigation


    // c'tor to create services
    override init() {
        super.init()

        print("Creating Location provider for HERE positioning...")

        do {
            self.engine = try LocationEngine()
        } catch let error {
            print("Error when creating HERE location engine: \(error)")
        }
    }

    
    func stopHEREPositioning() {
        guard engine != nil else { return }

        if engine!.isStarted {
            print("Stopping HERe positioning....")
            engine?.removeLocationDelegate(locationDelegate: self)
            engine?.removeLocationStatusDelegate(locationStatusDelegate: self)
            platformLocationService.delegate = nil
            engine?.stop()
        }
    }


    func startHEREPositioning() {
        guard engine != nil else { return }

        if engine!.isStarted == false {
            print("Starting HERE Location engine....")

            print("Attaching positioning delegates...")
            self.engine?.addLocationStatusDelegate(locationStatusDelegate: self)
            self.engine?.addLocationDelegate(locationDelegate: self)

            // listen for platform positioning updates, used for authorization only
            print("Attaching authorization delegates...")
            self.platformLocationService.delegate = self

            switch platformLocationService.authorizationStatus {
             case .notDetermined:
                 // Request permission if not determined yet
                 print("Authorization status unknown....Request authorization from user...")
                 self.platformLocationService.requestAlwaysAuthorization()
                 break

             case .restricted, .denied:
                 // Don't activate location features, since it was not allowed
                 print("Positioning denied by user already...")
                 break

             case .authorizedWhenInUse, .authorizedAlways:
                 // Enable location features after user allowed it
                 print("Positioning already authorized by user. Starting engine..")
                 _ = engine?.start(locationAccuracy: locationAccuracy)
                 break

             default:
                 print("Warning: Authorization without decision.")
                 break
             }

        } else {
            print("Warning: Position engine already started,...")
        }
    }


    // bunch of methods to control background location mode
    func setBackgroundMode(active: Bool) {
        // If plist entry for background location updates is active, the should not be needed anymore since it's also active by default then.
        // Anyway, doing this explicitly is good to show how to enable/disable it programatically on the fly.
        // See apple docs: https://developer.apple.com/documentation/corelocation/getting_the_user_s_location/handling_location_events_in_the_background
        _ = self.engine?.setBackgroundLocationIndicatorVisible(visible: true)       // forward to https://developer.apple.com/documentation/corelocation/cllocationmanager/2923541-showsbackgroundlocationindicator
        _ = self.engine?.setBackgroundLocationAllowed(allowed: active)              // forward to https://developer.apple.com/documentation/corelocation/cllocationmanager/1620568-allowsbackgroundlocationupdates
        _ = self.engine?.setPauseLocationUpdatesAutomatically(allowed: !active)     // forward to https://developer.apple.com/documentation/corelocation/cllocationmanager/1620553-pauseslocationupdatesautomatical
    }


    // @implements CLLocationManagerDelegate
    func locationManager(_ manager: CLLocationManager, didChangeAuthorization status: CLAuthorizationStatus) {

        switch status {
        case .restricted, .denied: // Disable location features
            print("Positioning denied by user after asking...")
            engine?.stop()
            break

        case .authorizedWhenInUse, .authorizedAlways: // Enable location features
            print("Positioning authorized by user after asking...")
            _ = engine?.start(locationAccuracy: locationAccuracy)
            break

        default:
            print("Closed authorizaiton dialog")
            break
        }
    }


    //@implements LocationStatusDelegate
    func onStatusChanged(locationEngineStatus: LocationEngineStatus) {
        print("Location status changed to \(locationEngineStatus.name)")

        switch locationEngineStatus {
            case .engineStarted:
                break
            case .engineStopped:
                break
            default:
                break
        }
    }


    //@implements LocationStatusDelegate
    func onFeaturesNotAvailable(features: [LocationFeature]) {
        for f in features {
            print("Feature not available: \(f)")
        }
    }


    //@implements LocationDelegate
    func onLocationUpdated(_ location: Location) {

        if navigationDelegate != nil {
            navigationDelegate?.onLocationUpdated(location)
        }

        if appDelegate != nil {
            appDelegate?.onLocationUpdated(location)
        }
    }

}
