//
//  LaneInformationView.swift
//  iOS Navigate Reference App
//
//  Created by Jonda, Sachin on 10.05.23.
//  Copyright © 2020 HERE Technologies. All rights reserved.
//

import SwiftUI
import heresdk

struct LaneInformationView: View {

    var nextLaneInformation = [heresdk.Lane]()

    func getRecommendationColor(for lane: Lane) -> SwiftUI.Color {
        let recommendationState = lane.recommendationState

        var highlightColor = SwiftUI.Color(#colorLiteral(red: 0, green: 0, blue: 0, alpha: 1))  // black as default if state undefined

        if recommendationState == .highlyRecommended {
            highlightColor = SwiftUI.Color(#colorLiteral(red: 0.4, green: 0.6, blue: 0.2, alpha: 1))  // dark green
        } else if recommendationState == .recommended {
            highlightColor = SwiftUI.Color(#colorLiteral(red: 0.6000000238, green: 0.6000000238, blue: 0.6000000238, alpha: 1))  // light gray
        }  else if recommendationState == .notRecommended {
            highlightColor = SwiftUI.Color(#colorLiteral(red: 0.8, green: 0.2, blue: 0.1, alpha: 1))  // red
        }

        return highlightColor
    }


    func getDirections(for lane: Lane) -> [String] {
        let directionCategory = lane.directionCategory

        var directions = [String]()

        // if lanes are access restricted, handle special cases
        if lane.access.automobiles == false && lane.access.buses == true {  // only busses, no cars
            directions.append("hds_lanetype_bus")
            return directions  // don't check for further types or direction information, since a icon is enough for me in this case
        }

        if lane.access.automobiles == false && lane.access.motorcycles == true {  // only scooters / motorcycles but no cars
            directions.append("hds_lanetype_motorscooter")
            return directions  // don't check for further types or direction information, since a icon is enough for me in this case
        }

        // special lane handling, difference to access restriction: different geometry of lanes
        if lane.type.isRegular == false {
            // comment: although special lany types can be combined, I always only show the most important icon and not multiple ones
            // commente#2: I also don't have assets for all cases, so i differentiate only first 3 cases I want to show the icons and let the other fall through
            // comment#3: Even special lanes can have directions and you might not tag them with icons, e.g. turn lanes.
            //            I prefer to show directions and no special icon for most of those special lanes
            if lane.type.isBicycle {
                directions.append("hds_lanetype_bicycle")
                return directions  // don't check for further types or direction information, since a icon is enough for me in this case
            } else if lane.type.isReversible {
                directions.append("hds_lanetype_reverse")
                return directions // don't check for further types or direction information, since a icon is enough for me in this case
            } else if lane.type.isHighOccupancyVehicle {
                directions.append("hds_lanetype_hov")
                return directions // don't check for further types or direction information, since a icon is enough for me in this case
            }

            /*
             // some more special lanes can be handled in future, not doing this right now. e.g.:
             else if lane.type.isExpress {
                directions.append("hds_lanetype_hov")
            } else if lane.type.isAcceleration {
                directions.append("hds_lanetype_hov")
            } else if lane.type.isDeceleration {
                directions.append("hds_lanetype_hov")
            } else if lane.type.isAuxiliary {
                directions.append("hds_lanetype_hov")
            } else if lane.type.isSlow {
                directions.append("hds_lanetype_hov")
            } else if lane.type.isShoulder {
                directions.append("hds_lanetype_hov")
            } else if lane.type.isRegulatedAccess {
                directions.append("hds_lanetype_hov")
            } else if lane.type.isTurn {
                directions.append("hds_lanetype_hov")
            } else if lane.type.isCenterTurn {
                directions.append("hds_lanetype_hov")
            } else if lane.type.isTruckParking {
                directions.append("hds_lanetype_hov")
            } else if lane.type.isParking {
                directions.append("hds_lanetype_hov")
            } else if lane.type.is_variable_driving {
                directions.append("hds_lanetype_hov")
            }
            */
        }

        // check for regular lane directions and show arrows

        if directionCategory.straight {
            directions.append("laneinfo_1")
        }

        if directionCategory.slightlyLeft {
            directions.append("laneinfo_2")
        }

        if directionCategory.quiteLeft {
            directions.append("laneinfo_4")
        }

        if directionCategory.hardLeft {
            directions.append("laneinfo_8")
        }

        if directionCategory.slightlyRight {
            directions.append("laneinfo_32")
        }

        if directionCategory.quiteRight {
            directions.append("laneinfo_64")
        }

        if directionCategory.hardRight {
            directions.append("laneinfo_128")
        }

        // add uturn lane arrow ONLy if no other arrows are shown.
        // normally uturn in combination with other arrows are a mess and not needed, so we don't show uturn arrow, if other arrows are there already
        if directions.isEmpty {
            if directionCategory.uTurnLeft {
                directions.append("laneinfo_16")
            } else if directionCategory.uTurnRight {
                directions.append("laneinfo_256")
            }
        }

        // add generic icon as fallback if error occured and no lane direction information is available
        // and also no lane type could be identified.
        if directions.isEmpty {
            directions.append("hds_lanetype_unknown")
        }

        return directions
    }


    var body: some View {
        GeometryReader { geometry in
            HStack(alignment: .center, spacing: 3) {
                Spacer()

                // for each lane evaluate image and recommendation
                ForEach(self.nextLaneInformation, id: \.self) { lane in

                    let dirs = getDirections(for: lane)

                    ZStack {  // overlay all arrow images
                        ForEach(dirs, id: \.self) { arrow in  // for each arrow on one lane sho a image overlay
                                Image( arrow )
                                    .resizable()
                                    .padding(.all, 10)
                                    .foregroundColor(SwiftUI.Color(#colorLiteral(red: 1, green: 1, blue: 1, alpha: 1)))
                                    .background(Color.clear)
                                    .frame(width: 50)
                        }
                    }
                    .background( lane.type.isBicycle ? SwiftUI.Color(#colorLiteral(red: 0.2392156869, green: 0.6745098233, blue: 0.9686274529, alpha: 1)) : getRecommendationColor(for: lane) )  // bicycle lanes always blue, other lanes based on recommended state

                }

                Spacer()
            }

        }
    }

}

struct LaneInformationView_Previews: PreviewProvider {
    static var previews: some View {
        LaneInformationView()
    }
}
