//
//  MapController.swift
//  iOS Navigate Reference App
//
//  Created by Jonda, Sachin on 20.02.22.
//  Copyright © 2023 HERE Technologies. All rights reserved.
//

import Foundation
import heresdk
import CoreLocation
import UIKit
import SwiftUI
import AVFoundation
import Combine

class MapController :
    NSObject,
    ObservableObject,
    AVSpeechSynthesizerDelegate,         // delegate for the TTS events
    LongPressDelegate,                  // gesture delegate
    TapDelegate,                        // gesture delegate
    PanDelegate,                        // gesture delegate
    RouteDeviationDelegate,             // navigation delegate
    RouteProgressDelegate,              // navigation delegate
    NavigableLocationDelegate,          // navigation delegate
    ManeuverNotificationDelegate,       // navigation delegate
    DestinationReachedDelegate,         // navigation delegate
    MilestoneStatusDelegate,            // navigation delegate
    SpeedWarningDelegate,               // navigation delegate
    SpeedLimitDelegate,                 // navigation delegate
    ManeuverViewLaneAssistanceDelegate, // navigation delegate
    JunctionViewLaneAssistanceDelegate, // navigation delegate
    SafetyCameraWarningDelegate,        // navigation delegate
    RoadTextsDelegate,                  // Navigation delegate
    RoadAttributesDelegate,             // Navigation delegate
    RoadSignWarningDelegate,            // Naviagtion delegate
    RealisticViewWarningDelegate,       // Naviagtion delegate
    LocationDelegate                    // get updates from LocationProviders
{
        
    private weak var mapView: MapView?

    // UI labels
    @Published var currentSpeed: Double = 0
    @Published var currentStreet = "unnamed"
    @Published var isSpeeding = false
    @Published var nextLaneInformation = [Lane]()

    @Published var currentSpeedLimit: Double?

    @Published var distanceToDestination = 0
    @Published var distanceToManeuver = 0
    @Published var timeToDestinationInSeconds = 0

    @Published var nextManeuverIcon = Utils.getIconNameforManeuver(action: ManeuverAction.continueOn) // default for now
    @Published var nextStreet = "unnamed"
    @Published var nextNextManeuverIcon: String?
    @Published var nextNextStreet = "unnamed"
    @Published var nextExitNr: String?

    // status flags for UI control
    @Published var isRouteAvailable = false
    @Published var isNavigationInProgress = false

    @Published var isAlertShown = false
    @Published var alertMessage = ""

    @Published var isWaitingDialogShown = false
    @Published var waitingDialogText = ""

    // generic attention dialog, mainly used for warnings to user/(driver, e.g. safety cameras
    @Published var attentionDialog: String?
    
    // junction view and sign post generated via SVG
    @Published var signPost: String?
    @Published var junctionView: String?
    
    // map items
    private var currentDestination: GeoCoordinates? // TODO: workaround for missing waypoints list in Route ?!
    private var activeLongPressMarker: MapMarker?
    private var activeRoutePolyline: MapPolyline?
    private var alternativeRoutePolyline: MapPolyline?
    //private var trafficPolylines = [MapPolyline]()  // save the traffic on route
    private var infoBubble: MapView.ViewPin?

    // routing engine (offline and online; sharing same interface)
    private var routingEngine: RoutingProtocol?

    // navigation members
    private var visualNavigator: VisualNavigator?
    private var activeRoute: Route?
    private var alternativeRoute: Route?
    private var lastManeuverIndex: Int32?

    private let REROUTE_THRESHOLD = 20.0; // necessary deviation in meters before re-routing is kicked off

    // default lookAt distances
    private let LOOKAT_DISTANCE_GLOBE:Double        = 15000000
    private let LOOKAT_DISTANCE_CONTINENTAL:Double  = 5000000
    private let LOOKAT_DISTANCE_COUNTRY:Double      = 1200000
    private let LOOKAT_DISTANCE_CITY:Double         = 50000
    private let LOOKAT_DISTANCE_DISTRICT:Double     = 1700

    // TTS Engine for iOS
    private lazy var speechSynthesizer = AVSpeechSynthesizer()

    private var offlineModeNotifier: AnyCancellable?
    
    // route colors
    private static let ACTIVE_COLOR = #colorLiteral(red: 0.2, green: 0.3, blue: 0.9, alpha: 1)
    private static let ALTERNATIVE_COLOR = #colorLiteral(red: 0.5, green: 0.9, blue: 0.9, alpha: 1) //Alpha for polylines not working:

    /*
    private static let JAMCOLORS = [ UIColor(red: 0.0, green: 0.9, blue: 0.0, alpha: 1.0),  // green
                                     UIColor(red: 0.0, green: 0.9, blue: 0.0, alpha: 1.0),  // green
                                     UIColor(red: 0.5, green: 0.9, blue: 0.0, alpha: 1.0),  // yellow-green
                                     UIColor(red: 0.5, green: 0.9, blue: 0.0, alpha: 1.0),  // yellow-green
                                     UIColor(red: 0.9, green: 0.9, blue: 0.0, alpha: 1.0),  // yellow
                                     UIColor(red: 0.9, green: 0.9, blue: 0.0, alpha: 1.0),  // yellow
                                     UIColor(red: 0.9, green: 0.5, blue: 0.0, alpha: 1.0),  // orange
                                     UIColor(red: 0.9, green: 0.5, blue: 0.0, alpha: 1.0),  // orange
                                     UIColor(red: 0.9, green: 0.0, blue: 0.0, alpha: 1.0),  // red
                                     UIColor(red: 0.9, green: 0.0, blue: 0.0, alpha: 1.0),  // red
                                     UIColor(red: 0.0, green: 0.0, blue: 0.0, alpha: 1.0)]  // black
    */

    // positioning indicator on map
    private var locIndicator3D = LocationIndicator()
    private var locIndicatorRaw: MapPolygon?  // debug visualization of raw position

    // positioning service by HERE
    private var hLocService: HERELocationProvider?

    // indicates if we should center map after position update. we normally only want this for the first update we receive
    @Published var centerOnPositionUpdate = true

    // simluation for positioning
    private var locationSimulator: LocationSimulator?
    private var isNavigationSimulated = false

    // create a GPX Logger to save GPX traces of NAvigation sessions
    private lazy var gpxLogger = GPXRecorder()

    // search
    private var searchEngine: SearchEngine?
    private var offlineSearchEngine: OfflineSearchEngine?

    private var searchResultCluster = MapMarkerCluster( imageStyle: MapMarkerCluster.ImageStyle(image:
                                                        MapImage( pixelData: #imageLiteral(resourceName: "Cluster Marker")
                                                                    .resizeImageWith(newSize: CGSize(width: 50,height: 50))
                                                                    .pngData()!, imageFormat: ImageFormat.png)) )

    private var currentSearchResultMarker = [MapMarker]()
    @Published var currentSuggestionsList = [Suggestion]()  // used in search bar drop down

    // general settings
    private var mapSettings: MapSettings!
    
    // offline map services: updater and downloader
    var mapLoaderController: MapLoaderController?    
    var mapUpdaterController: MapUpdaterController?

    // overload c'tor
    init(settings: MapSettings) {
        super.init()
        
        self.mapSettings = settings

        // listen for TTS events
        speechSynthesizer.delegate = self

       initOfflineServices()
    }
    
    
    private func initOfflineServices() -> Void {
        self.mapUpdaterController = MapUpdaterController(mapSettings: self.mapSettings)
        self.mapLoaderController = MapLoaderController(mapSettings: self.mapSettings)
    }


    private func getResourceStringFromBundle(name: String, type: String) -> String {
        let resourceUrl = Bundle.main.url(forResource: name, withExtension: type)
        
        guard let resourceString = resourceUrl?.path else {
            fatalError("Error: Resource not found!")
        }

        return resourceString
    }
    
    
    // initial map setup
    public func setupMap(mapView: MapView?, mapSettings: MapSettings) -> Void {

        let startTime = CFAbsoluteTimeGetCurrent() // measure time for seting up map

        guard mapView != nil else {
            print("MapView is not valid, cannot set up MapController!")
            return
        }

        self.mapView = mapView
        self.mapSettings = mapSettings
        
        
        // check and set if we are should be in offline mode.
        //SDKNativeEngine.sharedInstance?.setOfflineMode(offlineMode: self.mapSettings.offlineMode)
        //print("Offline mode status is: \(self.mapSettings.offlineMode)")

        // We need to listen for chnage event pusblished from settings to switch offline mode on and off during runtime
        // During setup, initial value will be sent once, and then afterwards subscribed for updates.
        // The assignment to the AnyCancellable instance keeps the lambda alive
        self.offlineModeNotifier = self.mapSettings.$offlineMode.sink { [weak self] offlineActivated in
            print("Offline mode changed to \(offlineActivated)")
        
            SDKNativeEngine.sharedInstance?.isOfflineMode = offlineActivated
            
            // When offline mode if activated, make sure also search and routing is switched
            self?.mapSettings.useOfflineSearch = offlineActivated
            self?.mapSettings.useOfflineRouting = offlineActivated
        }

        
        MapView.primaryLanguage = self.mapSettings.localeDeviceLanguageMapped // set map display language to device language - if available

        //let jsonResourceString = getResourceStringFromBundle(name: "MatrixStyle", type: "json")
        //self.mapView?.mapScene.loadScene(fromFile: jsonResourceString, completion: { (loadSceneError) in
        self.mapView?.mapScene.loadScene(mapScheme: self.mapSettings.mapScheme, completion: { (loadSceneError) in

            if loadSceneError != nil {
                print("Map scene loaded with error: \(String(describing: loadSceneError)) !")
            }

            // For debug purposed, print currently used accessKeyID and device language
            print("Using credentials: ID--> \(String(describing: SDKNativeEngine.sharedInstance?.options.accessKeyId))")
            print("Device language: LOCALE--> \(self.mapSettings.localeDeviceLanguage)")
            print("Screen scale factor: \(self.mapView?.contentScaleFactor ?? 0)")

            // set a default location for startup - will normallly overridden with the first positioning update
            self.mapView?.camera.lookAt(point: GeoCoordinates(latitude: 52.518043, longitude: 13.405991), orientation: GeoOrientationUpdate(bearing: 0, tilt: 0), zoom: MapMeasure(kind: .distance, value: self.LOOKAT_DISTANCE_GLOBE))

            self.mapView?.gestures.longPressDelegate = self
            self.mapView?.gestures.tapDelegate = self
            self.mapView?.gestures.panDelegate = self

            // position logo on bottom right corner
            self.mapView?.setWatermarkLocation(anchor: Anchor2D(horizontal: 1, vertical: 1) , offset: Point2D(x: -60*(self.mapView?.contentScaleFactor ?? 0),y: -60*(self.mapView?.contentScaleFactor ?? 0)) )

            // load map features from settings and display when needed
            self.loadMapFeatures()
            
            /*
            // enable just some categoriesd for embedded markers
            MapScene.setPoiVisibility(categoryIds: [PlaceCategory.eatAndDrink,
                                                    PlaceCategory.transport,
                                                    PlaceCategory.shopping,
                                                    PlaceCategory.accomodation,
                                                    PlaceCategory.sightsLandmarkAttaction,
                                                    PlaceCategory.facilities,
                                                    PlaceCategory.businessAndServices,
                                                    PlaceCategory.naturalAndGeographical,
                                                    PlaceCategory.areasAndBuildings],
                                      visibility: VisibilityState.hidden)
            */

            // init search
            do {
                self.searchEngine = try SearchEngine()
            } catch let error  {
                print("Error when creating search engine: \(error)")
                self.alertMessage = "ERROR: Could not create search engine: \(error)"
                self.isAlertShown = true
            }

            // enable HERE positioning
            self.hLocService = HERELocationProvider()
            guard self.hLocService != nil else {
                print("Warning: Could not start HERE location service")
                self.alertMessage = "Could not start HERE location service"
                self.isAlertShown = true
                return
            }
            self.hLocService?.appDelegate = self
            self.hLocService?.setBackgroundMode(active: false)
            self.hLocService?.startHEREPositioning()
            print("HERE location service started...")

            // setup location indicator
            self.locIndicator3D.locationIndicatorStyle  = .pedestrian
            self.locIndicator3D.isActive = true
            self.locIndicator3D.isAccuracyVisualized = true

            self.mapView?.addLifecycleDelegate(self.locIndicator3D)

            // if location service still holds a last known location, use this with grey marker as long as we don't have an updeate recent position
            if let last_known_location = self.hLocService?.engine?.lastKnownLocation {
                self.mapView?.camera.lookAt(point: last_known_location.coordinates, orientation: GeoOrientationUpdate(bearing: 0, tilt: 0), zoom: MapMeasure(kind: .distance, value: self.LOOKAT_DISTANCE_DISTRICT) )
                self.locIndicator3D.updateLocation(last_known_location.withoutAltitude())
            }

            // Mapsetup done
            let timeElapsed = CFAbsoluteTimeGetCurrent() - startTime
            print("Ready: Map set up within \(timeElapsed) seconds.")
        })
    }


    // @implements LocationDelegate for Location provider updates
    func onLocationUpdated(_ location: Location) {

        // draw/update raw position as a small blue dot
        let centerCircle = GeoCircle(center: location.coordinates, radiusInMeters: 1.0)
        let centerPolygon = GeoPolygon(geoCircle: centerCircle)

        // show raw position indicator - small blue dot
        if locIndicatorRaw == nil { // if not existing, create...
            locIndicatorRaw = MapPolygon(geometry: centerPolygon, color: #colorLiteral(red: 0.6000000238, green: 0.6000000238, blue: 0.6000000238, alpha: 1))
            locIndicatorRaw?.drawOrder = 1023 // try to put on top
            mapView?.mapScene.addMapPolygon(locIndicatorRaw!)
        } else {  // ...otherwise just update geometry/position
            locIndicatorRaw?.geometry = centerPolygon
        }

        // jump automatically to the first found position initially on position fix
        if self.centerOnPositionUpdate {
            self.mapView?.camera.lookAt(point: location.coordinates, orientation: GeoOrientationUpdate(bearing: 0, tilt: 0), zoom: MapMeasure(kind: .distance, value: self.LOOKAT_DISTANCE_DISTRICT))
            
            centerOnPositionUpdate = false // after we did this one, don't re-center anymore afterwards. also used in guidance to check if fix is available/first update came in
        }

        // update default position indicator with new position
        self.locIndicator3D.updateLocation(location.withoutAltitude())

        // save raw position in GPX logger if we are within navigation and recording is enabled
        if self.visualNavigator != nil &&  self.mapSettings.recordingActive == true && self.mapSettings.simulationMode == false {
            self.gpxLogger.addLocations(locations: [location])
        }

    }

    
    // re-center map to user position
    func zoomToMyPosition() -> Void {
        print("My position button clicked")

        guard hLocService != nil && hLocService?.engine != nil && hLocService?.engine?.lastKnownLocation != nil else {
            self.alertMessage = "Current position unknown. Please wait for positioning...."
            self.isAlertShown = true
            return
        }

        if self.isNavigationInProgress == false { // fly to current position
            let flyTo = MapCameraAnimationFactory.flyTo(target: GeoCoordinatesUpdate((hLocService?.engine?.lastKnownLocation!.coordinates)!),
                                                        orientation: GeoOrientationUpdate(GeoOrientation(bearing: 0, tilt: 0)),
                                                        zoom: MapMeasure(kind: .distance, value: self.LOOKAT_DISTANCE_DISTRICT),
                                                        bowFactor: 1,
                                                        duration: 1)
            mapView?.camera.startAnimation(flyTo)
            
        } else { // in navigation, just make sure we are back in camera tracking mode
                print("My position during navigation: enable following camera mode")
                self.visualNavigator?.cameraBehavior = DynamicCameraBehavior() // enable tracking the position due to customer interactiom
        }
    }


    // set map orientation back to north
      func northUp() -> Void {
        print("North up clicked")
        self.mapView?.camera.setOrientationAtTarget(GeoOrientationUpdate(bearing: 0, tilt: 0))
    }


    public func suggest(searchTerm: String) -> Void {
        
        if self.visualNavigator != nil {
                print("Please stop guidance first...")
                self.alertMessage = "Please stop guidance first before searching."
                self.isAlertShown = true
                return
        }

        guard self.mapView?.camera.state.targetCoordinates != nil else {
                print("Warning: No coordinates to search")
                self.alertMessage = "No coordinates to search"
                self.isAlertShown = true
                return
        }

        print("Query suggestions for \(searchTerm)")

        var options = SearchOptions()
        options.languageCode = self.mapSettings.localeDeviceLanguageMapped ?? LanguageCode.enUs
        options.maxItems = 15

        let tq = TextQuery(searchTerm, area: TextQuery.Area(areaCenter: (self.mapView?.camera.state.targetCoordinates)!) )

        if mapSettings.useOfflineSearch {
            // offline auto suggest
            _ = self.offlineSearchEngine?.suggest(textQuery: tq, options: options, completion: onSuggestCompleted)
        } else {
            // online auto suggest
            _ = self.searchEngine?.suggest(textQuery: tq, options: options, completion: onSuggestCompleted)
        }
    }


    private func onSuggestCompleted(error: SearchError?, resultSet: [Suggestion]?) {
        if resultSet != nil && resultSet!.count > 0 {
            print("Received suggestions: #\(String(describing: resultSet?.count))")
            self.currentSuggestionsList = resultSet!
        }
    }


    public func search(searchTerm: String) -> Void {

        if self.visualNavigator != nil {
            print("Please stop guidance first before searching...")
            self.alertMessage = "Please stop guidance first before searching..."
            self.isAlertShown = true
            return
        }

        guard self.mapView?.camera.state.targetCoordinates != nil else {
            print("Warning: No coordinates to search")
            self.alertMessage = "No coordinates to search"
            self.isAlertShown = true
            return
        }

        if searchTerm.isEmpty {
            print("Warning: Search term was empty")
            return
        }

        // show waiting dialog
        self.waitingDialogText = "Searching..."
        self.isWaitingDialogShown = true

        print("Searching for: \(searchTerm)")
        
        var options = SearchOptions()
        options.languageCode = self.mapSettings.localeDeviceLanguageMapped ?? LanguageCode.enUs

        let tq = TextQuery(searchTerm, area: TextQuery.Area(areaCenter: (self.mapView?.camera.state.targetCoordinates)!) )

        if mapSettings.useOfflineSearch == false { // if offline is not activated, go online
            _ = self.searchEngine?.search(textQuery: tq, options: options, completion: onSearchCompleted)
        } else {
            if self.offlineSearchEngine == nil {
                do {
                    self.offlineSearchEngine = try OfflineSearchEngine()
                } catch let error  {
                    print("ERROR: Could not create offline search engine: \(error)")
                    self.alertMessage = "ERROR: Could not create offline search engine: \(error)"
                    self.isAlertShown = true
                }
            }

            offlineSearchEngine?.search(textQuery: tq, options: options, completion: onSearchCompleted)
        }
    }


    private func onSearchCompleted(error: SearchError?, resultSet: [Place]?) {

        // always close search dialog
        self.isWaitingDialogShown = false

        if error != nil {
            print("Search returned with error \(String(describing: error?.description))")
            self.alertMessage = "\(error?.name ?? "Search problem occured.") "
            self.isAlertShown = true
            return
        }
        
        if resultSet != nil && resultSet!.count > 0 {
            print("Found search results: #\(String(describing: resultSet?.count))")

            // clear
            clearMap()

            for result in resultSet! {

                if result.geoCoordinates == nil {
                    print("Warning: Search result has no location attached!")
                    return
                }

                let mm = MapMarker( at: result.geoCoordinates!,
                                    image: MapImage( pixelData: #imageLiteral(resourceName: "Default POI Marker")
                                                        .resizeImageWith(newSize: CGSize(width: 50,height: 50))
                                                        .pngData()!, imageFormat: ImageFormat.png),
                                    anchor: Anchor2D(horizontal: 0.5,vertical: 1.0) )

                // add search results as payload into the mapmarker
                let md = Metadata()
                md.setString(key: "search_title", value: result.title)
                md.setString(key: "search_address", value: "\(result.address.street) \(result.address.houseNumOrName)" )
                md.setString(key: "search_category", value: result.details.categories.first?.id ?? "n/a")
                md.setString(key: "search_tel", value: (result.details.contacts.first?.landlinePhones.first?.phoneNumber ?? result.details.contacts.first?.mobilePhones.first?.phoneNumber ) ?? "n/a")

                mm.metadata = md

                self.currentSearchResultMarker.append( mm )
                searchResultCluster.addMapMarker(marker: mm)
            }

            self.mapView?.mapScene.addMapMarkerCluster(searchResultCluster)

            // Zoom to search results to give user a better understanding of the results on the map
            // if only one result was found, zoom to that (zoom to area only works if there are more than 1 result
            if resultSet?.count == 1 && resultSet?.first?.geoCoordinates != nil {
                let flyTo = MapCameraAnimationFactory.flyTo(target: GeoCoordinatesUpdate((resultSet?.first?.geoCoordinates)!),
                                                            orientation: GeoOrientationUpdate(GeoOrientation(bearing: 0, tilt: 0)),
                                                            zoom: MapMeasure(kind: .distance, value: self.LOOKAT_DISTANCE_DISTRICT),
                                                            bowFactor: 1,
                                                            duration: 1)
                mapView?.camera.startAnimation(flyTo)

            } else {
                var geoList = [GeoCoordinates]()
                _ = resultSet!.map {
                        if $0.geoCoordinates != nil {
                            geoList.append( $0.geoCoordinates! )
                        }
                    }

                if let searchResultBoundingBox = GeoBox.containing(geoCoordinates: geoList) {
                    let width = Double(UIScreen.main.bounds.width * UIScreen.main.nativeScale)
                    let height = Double(UIScreen.main.bounds.height * UIScreen.main.nativeScale)
                    let paddingRect = Rectangle2D( origin: Point2D(x: width*0.1, y: height*0.1), size: Size2D(width: width*0.8, height: height*0.8) ) // add some extra padding to the bounding box
                    self.mapView?.camera.lookAt(area: searchResultBoundingBox, orientation: GeoOrientationUpdate(bearing: Double.nan, tilt: Double.nan), viewRectangle: paddingRect)
                }
            }

        }

    }


    // start navigation, with or without route
    public func startNavigation(isSimulated: Bool) -> Void {
        guard mapView != nil else { return }

        guard self.hLocService != nil && self.hLocService?.engine != nil && self.hLocService?.engine?.isStarted  == true else {
            print("Warning: No positioning service is not running....")
            self.alertMessage = "No positioning service is running."
            self.isAlertShown = true
            return
        }

        guard self.hLocService?.engine?.lastKnownLocation != nil else {  // I also prevent on ui side start of guidance before position fix is there, but this is the safety net
            print("Warning: No position found.")
            self.alertMessage = "No current position found. Please wait for a position fix and green marker."
            self.isAlertShown = true
            return
        }

        if  isSimulated == true && self.activeRoute == nil // before starting all services, do some preconditionand sanity checks, e.g. simulation only with route
        {
            print("Warning, could not start simulation without route!")
            self.alertMessage = "Could not start simulation without route"
            self.isAlertShown = true
            return
        }

        self.isNavigationSimulated = isSimulated // save how guidance should be started

        print("Starting navigation...")

        // Create the Navigator....
        do {
            try self.visualNavigator = VisualNavigator()
        } catch let engineInstantiationError {
            print("Failed to initialize route navigator. Cause: \(engineInstantiationError)")
            self.alertMessage = "Failed to initialize route navigator. Cause: \(engineInstantiationError)"
            self.isAlertShown = true
            return
        }

        // ... and set it as delegate to the right location source (simulation or GNSS)
        if self.isNavigationSimulated == true {  // if in simulation mode....
            print("...with simulation")
            self.locationSimulator = try? LocationSimulator(route: self.activeRoute!, options: LocationSimulatorOptions(speedFactor: 1.2, notificationInterval: 0.5 ) )
            self.locationSimulator?.start()
            self.locationSimulator?.delegate = self.visualNavigator
        } else {  // no simulation, use real positioning to start guidance (tracking and route navigation supported)
            print("...with positioning")

            // enable background posiitoning if in real navigation
            self.hLocService?.setBackgroundMode(active: true)

            self.hLocService!.navigationDelegate = self.visualNavigator
        }

        // Debug print: show nr of available voice languages
        print("Checking languages supported: #\(String(describing: Navigator.availableLanguagesForManeuverNotifications().count))" )

        // connect the delegates
        self.visualNavigator?.navigableLocationDelegate = self
        self.visualNavigator?.routeProgressDelegate = self
        self.visualNavigator?.routeDeviationDelegate = self
        self.visualNavigator?.maneuverNotificationDelegate = self
        self.visualNavigator?.destinationReachedDelegate = self
        self.visualNavigator?.milestoneStatusDelegate = self
        self.visualNavigator?.speedWarningDelegate = self
        self.visualNavigator?.speedLimitDelegate = self
        self.visualNavigator?.maneuverViewLaneAssistanceDelegate = self
        self.visualNavigator?.junctionViewLaneAssistanceDelegate = self
        self.visualNavigator?.safetyCameraWarningDelegate = self
        self.visualNavigator?.roadTextsDelegate = self
        self.visualNavigator?.roadAttributesDelegate = self
        self.visualNavigator?.roadSignWarningDelegate = self
        self.visualNavigator?.realisticViewWarningDelegate = self   // SignPost + JunctionView = RealisticView

        // set a threshold for speed warning - in practice you don't wanna warn exactly on the edge, but have some tolerance.
        //self.guidanceNavigator?.speedWarningOptions = SpeedWarningOptions(speedLimitOffset: SpeedLimitOffset.init(lowSpeedOffsetInMetersPerSecond: 0.5, highSpeedOffsetInMetersPerSecond: 1.5, highSpeedBoundaryInMetersPerSecond: 1.5))

        // set voice guidance options like language
        //let mOpts = ManeuverNotificationOptions(language: LanguageCode.jaJp, unitSystem: UnitSystem.metric) // DEBUG for testing, set language hardcoded
        var mno = ManeuverNotificationOptions(language: Utils.getDeviceLanguageCode() ?? LanguageCode.enUs, unitSystem: UnitSystem.metric)
        mno.enableHighwayExit = true
        mno.enablePhoneme = false
        mno.includedNaturalGuidanceTypes = [.stopSign, .trafficLight]
        self.visualNavigator?.maneuverNotificationOptions = mno

        // Set tome realistifview options to configure it more for my usecase, like format choice
        self.visualNavigator?.realisticViewWarningOptions = RealisticViewWarningOptions(aspectRatio: .aspectRatio16X9, darkTheme: false)
        
        // don't register for all road signs here, but just some selected ones and some higher severity
        self.visualNavigator?.roadSignWarningOptions = RoadSignWarningOptions(typesFilter: RoadSignType.allCases,
                                                  categoriesFilter: [RoadSignCategory.warningSign],
                                                  generalWarningTypesFilter: GeneralWarningRoadSignType.allCases,
                                                  vehicleTypesFilter: RoadSignVehicleType.allCases)
        
        // Configure some colors
        //self.visualNavigator?.colors.maneuverArrowColor = UIColor.black
        //self.visualNavigator?.colors.setRouteProgressColors(sectionTransportMode: .truck, routeProgressColors: RouteProgressColors.init(ahead: UIColor.cyan, behind: UIColor.gray, outlineAhead: UIColor.black, outlineBehind: UIColor.white))
        
        // Setting route: it's route guidance.
        // Setting no route: it's tracking mode (no maneuver events, just progress update)
        if self.activeRoute != nil {
            self.visualNavigator?.route = self.activeRoute
            
            self.mapView?.mapScene.removeMapPolyline( self.activeRoutePolyline! )
            self.activeRoute = nil
            self.activeRoutePolyline = nil
            
            // if alternative routes are present, remove them from map (we don't want the user to select a new route during guidance)
            if(self.alternativeRoute != nil){
                self.mapView?.mapScene.removeMapPolyline( self.alternativeRoutePolyline! )
                self.alternativeRoutePolyline = nil
                self.alternativeRoute = nil
            }
        } else {
            // and in tracking mode, setup sepcial tracking options to indicate we are a truck and want truck warnings also in tracking mode (in route navigation, route indicates this)
            self.visualNavigator?.trackingTransportProfile = TransportProfile(vehicleProfile: VehicleProfile(vehicleType: (mapSettings.transportMode == TransportMode.truck ? VehicleType.truck : VehicleType.car) )) // TODO: Not nice to convert hier partially
        }

        // some UI tweaks for guidance:

        // adapt position of watermark -- TODO
        //self.mapView?.setWatermarkPosition(placement: WatermarkPlacement.bottomCenter,
        //                                   bottomCenterMargin: UInt32(10*self.mapView!.contentScaleFactor)+UInt32(100*self.mapView!.contentScaleFactor))  // dynamic spacing plus static height of bottom status bar for navigation (100ox for all resultions)

        // Visual guidance options
        self.visualNavigator?.isRouteVisible = true   // use route rendering an features from VN
        self.visualNavigator?.isRouteProgressVisible = true // use route eatup visualization
        
        self.visualNavigator?.cameraBehavior = DynamicCameraBehavior() // auto zoom

        // set mapview anchor points to lower quarter of map - in navigation you are less interested in the stuff behind you, but more looking ahead instead
        // I can also the visual navigator do that, but for my UI the point is too low
        self.visualNavigator?.cameraBehavior?.normalizedPrincipalPoint = Anchor2D(horizontal: 0.5,vertical: 0.7)

        self.visualNavigator?.startRendering(mapView: self.mapView!)
        
        // WORKAROUND: remove pedestrian default position indicator since VN brings own one
        self.mapView?.removeLifecycleDelegate(self.locIndicator3D)

        self.isNavigationInProgress = true

        print("Navigation setup done...")
    }


    // stop navigation
    public func stopNavigation() -> Void {
        print("Stopping navigation....")

        if self.isNavigationSimulated {
            self.locationSimulator?.stop()
            self.locationSimulator?.delegate = nil
            self.locationSimulator = nil

            self.isNavigationSimulated = false
        }

        if self.visualNavigator != nil {
            self.visualNavigator?.stopRendering() // TBD: do I need to reset all VN options too ?

            self.visualNavigator?.cameraBehavior = nil
            
            // WORKAROUND: re-add pedestrian default position indicator since VN brings own one
            self.mapView?.addLifecycleDelegate(self.locIndicator3D)

            self.visualNavigator?.routeProgressDelegate = nil
            self.visualNavigator?.destinationReachedDelegate = nil
            self.visualNavigator?.milestoneStatusDelegate = nil
            self.visualNavigator?.routeDeviationDelegate = nil
            self.visualNavigator?.navigableLocationDelegate = nil
            self.visualNavigator?.maneuverNotificationDelegate = nil
            self.visualNavigator?.speedWarningDelegate = nil
            self.visualNavigator?.speedLimitDelegate = nil
            self.visualNavigator?.maneuverViewLaneAssistanceDelegate = nil
            self.visualNavigator?.junctionViewLaneAssistanceDelegate = nil
            self.visualNavigator?.safetyCameraWarningDelegate = nil
            self.visualNavigator?.roadTextsDelegate = nil
            self.visualNavigator?.roadAttributesDelegate = nil
            self.visualNavigator?.roadSignWarningDelegate = nil
            self.visualNavigator?.realisticViewWarningDelegate = nil

            self.visualNavigator?.route = nil

            if self.hLocService != nil {
                self.hLocService?.setBackgroundMode(active: false) // outside of navigation, disable background mode
                self.hLocService!.navigationDelegate = nil // remove also navigator as delegate from provider since we set this explicitly before
            }

            self.visualNavigator = nil
        }

        // reset event data
        self.isSpeeding = false

        self.currentSpeed = 0
        self.currentSpeedLimit = nil
        self.currentStreet = "unnamed"

        self.nextManeuverIcon = Utils.getIconNameforManeuver(action: ManeuverAction.continueOn) // default for now
        self.nextStreet = "unnamed"
        self.nextNextManeuverIcon = nil
        self.nextNextStreet = "unnamed"

        self.nextLaneInformation.removeAll() // delete all information about lanes

        self.distanceToDestination = 0
        self.distanceToManeuver = 0
        self.timeToDestinationInSeconds = 0
        
        self.attentionDialog = nil
        self.junctionView = nil
        self.signPost = nil

        // save the GPX trace, if enabled in menu and we are not in simulation
        if self.mapSettings.recordingActive == true && self.mapSettings.simulationMode == false {
            self.gpxLogger.saveRecording()
        }

        // reset position of watermark
        // self.mapView?.setWatermarkPosition(placement: WatermarkPlacement.bottomCenter, bottomCenterMargin: UInt32(10*self.mapView!.contentScaleFactor))

        self.isNavigationInProgress = false

        clearMap()

        // when guidance stops, keep current position and zoom out
        self.mapView?.camera.lookAt( point: (mapView?.camera.state.targetCoordinates)!, orientation: GeoOrientationUpdate(bearing: 0, tilt: 0),zoom: MapMeasure(kind: .distance, value: self.LOOKAT_DISTANCE_DISTRICT) )
    }


    // @implement Maneuevr for navigation
    func onManeuverNotification(_ text: String) {

        // Take over audio stream control to mix/duck with music if needed
        // For audio setting proposales by Apple/CarPlay see:
        // https://developer.apple.com/carplay/documentation/CarPlay-Navigation-App-Programming-Guide.pdf#page=33

        do {
            try AVAudioSession.sharedInstance().setActive(true)
        } catch {
            print("Warning: Could not take over audio stream for TTS control")
        }

        do {
            try AVAudioSession.sharedInstance().setCategory(AVAudioSession.Category.playback, options: [.duckOthers, .interruptSpokenAudioAndMixWithOthers])
            try AVAudioSession.sharedInstance().setMode(.voicePrompt) // only available after playback category has been set before. see carplay doc mentioned above for details.
        } catch {
            print("Warning: Could not set configuration/mode for TTS on current audio stream: \(error)")
        }

        print("TTS Text: \(text), using language: \(String(describing: (self.mapSettings.localeDeviceLanguageMapped != nil) ? self.mapSettings.localeDeviceLanguage : "en-US")) ")

        let speechUtterance: AVSpeechUtterance = AVSpeechUtterance(string: text)
        //speechUtterance.voice = AVSpeechSynthesisVoice(language: "ja-JP" ) // DEBUG: hardcode language for testing
        speechUtterance.voice = AVSpeechSynthesisVoice(language: (self.mapSettings.localeDeviceLanguageMapped != nil) ? self.mapSettings.localeDeviceLanguage : "en-US" ) // use device language if supported by navigation, otherwise fallback to US english

        speechSynthesizer.speak(speechUtterance)
    }

    
    // @implement AVSpeechSynthesizerDelegate
    func speechSynthesizer(_ synthesizer: AVSpeechSynthesizer, didFinish utterance: AVSpeechUtterance){
        print("TTS Audio done.")

        do {
            try AVAudioSession.sharedInstance().setActive(false, options: [.notifyOthersOnDeactivation])
        } catch {
            print("Warning: Could not return audio stream control.")
        }
    }

    // @implement NavigableLocationDelegate for navigation
    func onNavigableLocationUpdated(_ navigableLocation: NavigableLocation) {

        if navigableLocation.mapMatchedLocation?.coordinates == nil {
            print("Warning: Could not map match position. You are off track. Using raw position")
        }

        self.currentSpeed = (navigableLocation.originalLocation.speedInMetersPerSecond ?? 0) * Utils.METER_PER_SECOND_TO_KMH
    }


    //@implement RoadTextDelegate
    func onRoadTextsUpdated(_ roadTexts: RoadTexts) {
        print("Road Text update: \(String(describing: roadTexts.names.defaultValue()))")

        // show proper naming information about current street
        // use from road attributes! TODO!
        //if roadTexts.roadType == RoadType.highway { // on highways, we want to show highway number instead of street, while inner city and urban areas, we shos street names of road numbers
            //self.currentStreet = roadTexts.numbers.defaultValue() ?? roadTexts.names.defaultValue() ?? "unnamed"
            self.currentStreet = roadTexts.names.defaultValue() ?? "unnamed"
        //} else {
        //    self.currentStreet = roadTexts.names.defaultValue() ?? nextManeuver?.roadTexts.numbers.defaultValue() ?? "unnamed"
        //}

    }

    //@implement RoadAttributeDelegate
    func onRoadAttributesUpdated(_ roadAttributes: RoadAttributes) {

            if (roadAttributes.isDirtRoad)
            {
                print("Road attribute: Dirty road")
            }

            if (roadAttributes.isTunnel)
            {
                print("Road attribute: Tunnel")
                self.attentionDialog = "Tunnel detected. Lights on?"
                Timer.scheduledTimer(withTimeInterval: 3.0, repeats: false) { timer in  // kill dialog again in few seconds
                    self.attentionDialog = nil
                }
            }

            if (roadAttributes.isBridge)
            {
                print("Road attribute: ridge")
            }

            if (roadAttributes.isRamp)
            {
                print("Road attribute: Ramp")
            }

            if (roadAttributes.isControlledAccess)
            {
                print("Road attribute: Highway / Controlled access road")
            }

            if (roadAttributes.isPrivate)
            {
                print("Road attribute: Private road")
            }

            if (roadAttributes.isNoThrough)
            {
                print("Road attribute: No throught / You can only enter if your destination is located there")
            }

            if (roadAttributes.isTollway)
            {
                print("Road attribute: Toll road")
            }

            if (roadAttributes.isDividedRoad)
            {
                print("Road attribute: Divided road")
            }

            if (roadAttributes.isRoundabout)
            {
                print("Road attribute: Roundabout")
            }

            if (roadAttributes.isRightDrivingSide)
            {
                print("Road attribute: Right hand driving side country")
            } else {
                print("Road attribute: Left hand driving side country")
            }
    }

    
    // @implement realisticViewWarningDelegate
    func onRealisticViewWarningUpdated(_ realisticViewWarning: heresdk.RealisticViewWarning) {
        print("JunctionView with SignPost \(realisticViewWarning.distanceType.name) in \(realisticViewWarning.distanceToRealisticViewInMeters) m!")

        if realisticViewWarning.distanceType == DistanceType.ahead {
            
            self.junctionView = realisticViewWarning.realisticView?.junctionViewSvgImageContent
            self.signPost = realisticViewWarning.realisticView?.signpostSvgImageContent
            
            guard self.junctionView != nil && self.signPost != nil else {
                print("Warning: RealtityView images not available in callback.")
                return
            }
            
            self.attentionDialog = "RealisticView \(realisticViewWarning.distanceType.name) in \(Utils.formatDistance(meter: Int(realisticViewWarning.distanceToRealisticViewInMeters)))"
            
            // DEBUG !!! Save all SVGs to disk!
            /*
            let jv_file = "\(NSDate().timeIntervalSince1970)_junctionView.svg"
            if let dir = FileManager.default.urls(for: .documentDirectory, in: .userDomainMask).first {
                let fileURL = dir.appendingPathComponent(jv_file)
                do {
                    try self.junctionView?.write(to: fileURL, atomically: false, encoding: .utf8)
                }
                catch {
                    print("ERROR: Could not save SVG file of JunctionView to disk.")
                }
            }

            // DEBUG !!! Save all SVGs to disk!
            let sp_file = "\(NSDate().timeIntervalSince1970)_signPostView.svg"
            if let dir = FileManager.default.urls(for: .documentDirectory, in: .userDomainMask).first {
                let fileURL = dir.appendingPathComponent(sp_file)
                do {
                    try self.signPost?.write(to: fileURL, atomically: false, encoding: .utf8)
                }
                catch {
                    print("ERROR: Could not save SVG file of JunctionView to disk.")
                }
            }
             */

            
        } else {
            self.attentionDialog = nil
            self.junctionView = nil
            self.signPost = nil
        }

    }
    
    
    // @implement RoadSignWarningDelegate
    func onRoadSignWarningUpdated(_ roadSignWarning: RoadSignWarning) {        
        print("Received road sign warning: \(roadSignWarning)")
                
        self.attentionDialog = "\(roadSignWarning.category.name) sign ahead in \(Utils.formatDistance(meter: Int(roadSignWarning.distanceToRoadSignInMeters)))"
        
        Timer.scheduledTimer(withTimeInterval: 5.0, repeats: false) { timer in  // kill dialog again in few seconds
            self.attentionDialog = nil
        }
    }
    
    
    // @implement RouteProgressDelegate
    func onRouteProgressUpdated(_ routeProgress: RouteProgress) {

        let nextManeuverProgress = routeProgress.maneuverProgress.first // next maneuever
        let nextNextManeuverProgress = routeProgress.maneuverProgress.count > 1 ? routeProgress.maneuverProgress[1]:nil // next-next maneuever

        guard nextManeuverProgress != nil else { return }  // make sure next maneuver is not nil

        // update all information about next maneuver
        let nextManeuver = visualNavigator?.getManeuver(index: Int32(nextManeuverProgress!.maneuverIndex))
        self.nextManeuverIcon = Utils.getIconNameforManeuver(action: nextManeuver?.action ?? ManeuverAction.continueOn)

        // DEBUG infos about maneuvers
        //print("Next maneuver is of type: \(nextManeuver?.roadType)")
        //print("Maneuver angle: \(nextManeuver?.turnAngleInDegrees)")
        //print("Roundabout angle: \(nextManeuver?.roundaboutAngleInDegrees)")

        // show proper naming for next maneuver and following street
        if nextManeuver?.nextRoadType == RoadType.highway {  // on highways, we want to show highway number instead of street, while inner city and urban areas, we shos street names of road numbers
            self.nextStreet = nextManeuver?.nextRoadTexts.numbersWithDirection.defaultValue() ?? nextManeuver?.nextRoadTexts.names.defaultValue() ?? "unnamed"
            // For highways, also attach 'towards' if available
            if let towardsText = nextManeuver?.signpost?.labels.first {
                self.nextStreet = self.nextStreet + " towards \(towardsText)"
            }
        } else {
            self.nextStreet = nextManeuver?.nextRoadTexts.numbersWithDirection.defaultValue() ?? nextManeuver?.nextRoadTexts.numbersWithDirection.defaultValue() ?? "unnamed"
        }

        // expose exit sign information, available normally only on highways
        self.nextExitNr = nextManeuver?.exitSignTexts.defaultValue()

        if self.nextExitNr != nextManeuver?.exitSignTexts.defaultValue() {
            print("Take Exit #: \(nextManeuver?.exitSignTexts.defaultValue() ?? "n/a")")
        }

        // check if we are on our final street before arrival and how arrival information instead
        if nextManeuver?.action == ManeuverAction.arrive {
            self.nextStreet = "Destination ahead"
        }

        // same naming for nextNext maneuver
        if nextNextManeuverProgress != nil {
            let nextNextManeuver = visualNavigator?.getManeuver(index: Int32(nextNextManeuverProgress!.maneuverIndex))
            self.nextNextManeuverIcon = Utils.getIconNameforManeuver(action: nextNextManeuver?.action ?? ManeuverAction.continueOn)

            if nextNextManeuver?.nextRoadType == RoadType.highway {  // on highways, we want to show highway number instead of street, while inner city and urban areas, we shos street names of road numbers
                self.nextNextStreet = nextNextManeuver?.nextRoadTexts.numbersWithDirection.defaultValue() ?? nextNextManeuver?.nextRoadTexts.names.defaultValue() ?? "unnamed"
                // For highways, also attach 'towards' if available
                if let towardsText = nextNextManeuver?.signpost?.labels.first {
                    self.nextNextStreet = self.nextNextStreet + " towards \(towardsText)"
                }
            } else {
                self.nextNextStreet = nextNextManeuver?.nextRoadTexts.names.defaultValue() ?? nextNextManeuver?.nextRoadTexts.numbersWithDirection.defaultValue() ?? "unnamed"
            }

            // check if we are on our final street before arrival and how arrival information instead
            if nextNextManeuver?.action == ManeuverAction.arrive {
                self.nextNextStreet = "Destination ahead"
            }
        } else {
            self.nextNextManeuverIcon = nil
            self.nextNextStreet = "unnamed"
        }

        // Update more UI Labels
        self.distanceToManeuver = Int(nextManeuverProgress!.remainingDistanceInMeters)

        /*
        // show proper naming information about current street
        if nextManeuver?.roadType == RoadType.highway { // on highways, we want to show highway number instead of street, while inner city and urban areas, we shos street names of road numbers
            self.currentStreet = nextManeuver?.roadTexts.numbers.defaultValue() ?? nextManeuver?.roadTexts.names.defaultValue() ?? "unnamed"
        } else {
            self.currentStreet = nextManeuver?.roadTexts.names.defaultValue() ?? nextManeuver?.roadTexts.numbers.defaultValue() ?? "unnamed"
        }
         */

        // routeProgress will be called multiple times for the same maneuver (e.g. for distance updates).
        // Here we check if some compeletely new maneuver came in, so we can reset some things for the nex maneuver
        if self.lastManeuverIndex != nextManeuverProgress!.maneuverIndex {
            let debugMessage = "Next maneuver is \(String(nextManeuver?.action.name ?? "n/a")) on \(String(describing: self.nextStreet)) in \(nextManeuverProgress!.remainingDistanceInMeters) meters. The next-next maneuver is in \(String(describing: nextNextManeuverProgress?.remainingDistanceInMeters)) meters."
            print("New maneuver: \(debugMessage)")

            // reset lane display if a new maneuver is coming
            self.nextLaneInformation.removeAll()
        }

        lastManeuverIndex = nextManeuverProgress!.maneuverIndex

        // update for dashoboard numbers
        self.distanceToDestination = Int(routeProgress.sectionProgress.last?.remainingDistanceInMeters ?? 0)
        self.timeToDestinationInSeconds = Int(routeProgress.sectionProgress.last?.remainingDuration ?? 0)
        
        // update route for route eatup rendering
        // not needed anymore, using route progress of VN
        //let progress_onRoute = 1-(Double(routeProgress.sectionProgress.last!.remainingDistanceInMeters)/Double(self.activeRoute!.lengthInMeters))
        //self.activeRoutePolyline?.progress = progress_onRoute
    }


    // @implement RouteDeviationDelegate
    func onRouteDeviation(_ routeDeviation: RouteDeviation) {

        let currentCoord = routeDeviation.currentLocation.mapMatchedLocation?.coordinates ?? routeDeviation.currentLocation.originalLocation.coordinates
        let lastCoordOnRoute = routeDeviation.lastLocationOnRoute?.mapMatchedLocation?.coordinates ?? routeDeviation.lastLocationOnRoute?.originalLocation.coordinates

        if lastCoordOnRoute != nil {
            let distanceInMeters = currentCoord.distance(to: lastCoordOnRoute!)
            print("RouteDeviation in meters is \(distanceInMeters)")

            if distanceInMeters < self.REROUTE_THRESHOLD {  // if deviation is less than threshold don't calculate new route
                return
            }
        }else {
            print("No last point on current route found")
        }

        guard self.currentDestination != nil else {
            print("ERROR: No last destination available!")
            self.alertMessage = "No last destination available"
            self.isAlertShown = true
            return
        }

        // re-calculate route from current posiution to last set destination
        self.calculateRoute(to: self.currentDestination!) { (error: RoutingError?, routes: [Route]?) in
                self.isWaitingDialogShown = false  // always cancel progress dialog

                guard error == nil else {
                    print("Error: Could not calculate route \(String(describing: error?.name))")
                    self.alertMessage = "Could not calculate route \(String(describing: error?.name))"
                    self.isAlertShown = true
                    return
                }

                print("New route calculated... Got : \(String(describing: routes?.count)) routes from system")

                self.addRoutesToMap(routes: routes, zoomToRoute: false)  // replace route on mapview but don't zoom in
                self.visualNavigator?.route = self.activeRoute  // set new route for guidance
        }
    }


    // @implement DestinationReachedDelegate
    func onDestinationReached() {
        print("Destination reached...Stopping Navigation")
        stopNavigation()
    }


    // @implement MilestoneStatusDelegate
    func onMilestoneStatusUpdated(milestone: Milestone, status: MilestoneStatus) {
        print("Navigation: Milestone #\(milestone.waypointIndex.debugDescription) status: \(status.name)")
    }


    //@implement SpeedLimitDelegate
    func onSpeedLimitUpdated(_ speedLimit: SpeedLimit) {

        // set new base speed limit, taking special speed limits that are effctive into account
        self.currentSpeedLimit = speedLimit.effectiveSpeedLimitInMetersPerSecond()

        // reset speed limit advisory dialog
        self.attentionDialog = nil

        // check for special speed limits and show warning dialog too
        if let specialSpeed = speedLimit.schoolZoneSpeedLimitInMetersPerSecond {
            self.attentionDialog = String(format: "School nearby: Reduce speed to %.0f km/h !", specialSpeed*Utils.METER_PER_SECOND_TO_KMH)
        }
        else if let specialSpeed = speedLimit.advisorySpeedLimitInMetersPerSecond  {
            self.attentionDialog = String(format: "Advisory: Reduce speed for safety reasons to %.0f km/h !", specialSpeed*Utils.METER_PER_SECOND_TO_KMH)
        }
        else if let specialSpeed = speedLimit.fogSpeedLimitInMetersPerSecond {
            self.attentionDialog = String(format: "If fog reduce speed to %.0f km/h !", specialSpeed*Utils.METER_PER_SECOND_TO_KMH)
        }
        else if let specialSpeed = speedLimit.snowSpeedLimitInMetersPerSecond {
            self.attentionDialog = String(format: "If snow reduce speed to %.0f km/h !", specialSpeed*Utils.METER_PER_SECOND_TO_KMH)
        }
        else if let specialSpeed = speedLimit.rainSpeedLimitInMetersPerSecond {
            self.attentionDialog = String(format: "If rain reduce speed to %.0f km/h !", specialSpeed*Utils.METER_PER_SECOND_TO_KMH)
        }
    }


    //@implement SpeedWarningDelegate
    func onSpeedWarningStatusChanged(_ status: SpeedWarningStatus) {
        print("Speeding detected...")

        if status == SpeedWarningStatus.speedLimitExceeded {
            self.isSpeeding = true
        } else {
            self.isSpeeding = false
        }
    }


    //@implement SafetyCameraWarningDelegate
    public func onSafetyCameraWarningUpdated(_ warning: SafetyCameraWarning) {
        print("Camera of type \(warning.type.name) detected ahead in \(warning.distanceToCameraInMeters)m. Remember, limit is: \(warning.speedLimitInMetersPerSecond*Utils.METER_PER_SECOND_TO_KMH) km/h")

        self.attentionDialog = "Camera ahead in \(Utils.formatDistance(meter: Int(warning.distanceToCameraInMeters)))"
        Timer.scheduledTimer(withTimeInterval: 3.0, repeats: false) { timer in  // kill dialog again in few seconds
            self.attentionDialog = nil
        }
    }


    //@implement LaneAssistanceDelegate
    func onLaneAssistanceUpdated(_ laneAssistance: ManeuverViewLaneAssistance) {
        print("Lane information received...\(laneAssistance.lanesForNextManeuver.count) lanes available for next maneuver")
        self.nextLaneInformation = laneAssistance.lanesForNextManeuver

        // Comment out debug if wanted - just debug print for lane assiatnce
        for s in laneAssistance.lanesForNextManeuver {
            print("State: \(s.recommendationState.description) - Type: \(s.type) - Cat: \(s.directionCategory)")
        }
    }

    
    // @implement JunctionAssistanceDelegate
    func onLaneAssistanceUpdated(_ laneAssistance: JunctionViewLaneAssistance) {
        print("Lane information received...\(laneAssistance.lanesForNextJunction.count) lanes available for next maneuver")
        self.nextLaneInformation = laneAssistance.lanesForNextJunction

        // Comment out debug if wanted - just debug print for lane assiatnce
        for s in laneAssistance.lanesForNextJunction {
            print("State: \(s.recommendationState.description) - Type: \(s.type) - Cat: \(s.directionCategory)")
        }

    }


    // @implement PanDelegate
    func onPan(state: heresdk.GestureState, origin: Point2D, translation: Point2D, velocity: Double) {
        if self.isNavigationInProgress {
            print("Panning during navigation: releaseing following camera mode")
            self.visualNavigator?.cameraBehavior = nil // disable tracking the position due to customer interactiomn
        }
    }


    // @implement TapDelegate
    func onTap(origin: Point2D) {

        // close infobubble on tap event
        if self.infoBubble != nil {
            self.infoBubble?.unpin()
            self.infoBubble = nil
        }

        // check if we picked a map item (marker or polyline) and trigger action
        self.mapView?.pickMapItems(at: origin, radius: 20.0, completion: onMapItemsPicked)

        // check if user clicked on embedded carto POI marker
        let pickingArea = Rectangle2D(origin: origin,size: Size2D(width: 10, height: 10))
        self.mapView?.pickMapContent(inside: pickingArea, completion: onMapCartoItemPicked)
    }


    // Completion handler for carto POI marker picking
    func onMapCartoItemPicked(pickedMapItems: PickMapContentResult?){

        if pickedMapItems?.pois.first != nil {

            //clear map
            clearMap()

            let topmostMapMarker = pickedMapItems!.pois.first  // get top most marker clicked (if piled)

            // wrap SwiftUI View in UIView for now, as long as there's no native SwiftUI support
            // => [IOTSDK-5815]
            let ib = InfoBubbleView(titleText: topmostMapMarker?.name ?? "n/a", addressText: "n/a", telText: "n/a")
            let infoBubbleView = UIHostingController(rootView: ib).view

            if infoBubbleView != nil {
                infoBubbleView?.sizeToFit() // kills transparancy but needed for anchor point
                infoBubbleView?.backgroundColor = UIColor.clear // set transparancy explicitely now again
                infoBubbleView?.layer.anchorPoint = CGPoint(x: 0.5, y: 1.0) // set anchorpoint bottom center to peak
                // Unsure if there's a better way to solve the transparancy issue after swiftui view to uiview conversion

                self.infoBubble = self.mapView?.pinView(infoBubbleView!, to: topmostMapMarker!.coordinates)
            }

            // place selected marker like long press POI
            self.activeLongPressMarker = MapMarker( at: topmostMapMarker!.coordinates,
                                                  image: MapImage( pixelData: #imageLiteral(resourceName: "Default POI Marker")
                                                                    .resizeImageWith(newSize: CGSize(width: 50,height: 50))
                                                                    .pngData()!, imageFormat: ImageFormat.png),
                                                  anchor: Anchor2D(horizontal: 0.5,vertical: 1.0) )
            self.mapView?.mapScene.addMapMarker( activeLongPressMarker! )

            // fetch more information from offline engine and update infobubble if result is available
            if topmostMapMarker?.offlineSearchId != nil {
                fetchCartoPOIDetails(offlineSearchId: topmostMapMarker!.offlineSearchId)
            }

            // calculate route to this picked item
            calculateAndAddRoutesToMap(to: topmostMapMarker!.coordinates)

        } else {
            print("No carto marker clicked")
        }
    }


    // Get details for a certain place(id) via offline engine if data is available (streamed or preloaded)
    // and replace info bubble with more detailed one
    func fetchCartoPOIDetails(offlineSearchId: String) {

        if self.offlineSearchEngine == nil {
            do {
                self.offlineSearchEngine = try OfflineSearchEngine()
            } catch let error  {
                print("ERROR: Could not create offline search engine: \(error)")
                self.alertMessage = "ERROR: Could not create offline search engine: \(error)"
                self.isAlertShown = true

                return
            }
        }

        let lang = self.mapSettings.localeDeviceLanguageMapped ?? LanguageCode.enUs

        self.offlineSearchEngine?.search(placeIdQuery: PlaceIdQuery(offlineSearchId), languageCode: lang) { error, place in
            print("Place details found: \(place?.title ?? "<not found>") at \(place?.address.addressText ?? "n/a")")

            // update bubble if we found some result with coordinates and title, and the infobubble is still shown
            if self.infoBubble != nil && place?.title != nil && place?.geoCoordinates != nil {

                // wrap SwiftUI View in UIView for now, as long as there's no native SwiftUI support => [IOTSDK-5815]
                let ib = InfoBubbleView(titleText: place?.title ?? "n/a",
                                        addressText: place?.address.addressText ?? "n/a",
                                        telText: place?.details.contacts.first?.landlinePhones.first?.phoneNumber ?? place?.details.contacts.first?.mobilePhones.first?.phoneNumber ?? "n/a")
                let infoBubbleView = UIHostingController(rootView: ib).view
                infoBubbleView?.sizeToFit() // kills transparancy but needed for anchor point
                infoBubbleView?.backgroundColor = UIColor.clear // set transparancy explicitely now again
                infoBubbleView?.layer.anchorPoint = CGPoint(x: 0.5, y: 1.0) // set anchorpoint bottom center to peak
                // Unsure if there's a better way to solve the transparancy issue after swiftui view to uiview conversion

                self.infoBubble?.unpin() // unpin old one first
                self.infoBubble = self.mapView?.pinView(infoBubbleView!, to: place!.geoCoordinates!) // and add updated one
            }
        }
    }


    // Completion handler to receive picked map items.
    func onMapItemsPicked(pickedMapItems: PickMapItemsResult?) {

        guard pickedMapItems != nil else {
            print("No map item picked")
            return
        }

        // check for normal mao marker picked or cluster marker was picked
        if pickedMapItems?.markers.first != nil || pickedMapItems?.clusteredMarkers.first?.markers.first != nil {

            //clear map
            clearMap()

            let topmostMapMarker = pickedMapItems?.markers.first ?? pickedMapItems?.clusteredMarkers.first?.markers.first

            // if search result picked, we have a metadata
            let titleText = topmostMapMarker!.metadata?.getString(key: "search_title")
            let addressText = topmostMapMarker!.metadata?.getString(key: "search_address")
            let telText = topmostMapMarker!.metadata?.getString(key: "search_tel")
            print("Clicked on map marker: \(String(describing: titleText))")

            // Show an infobubble on clicked marker
            // only for search results, not long press marker
            if titleText != nil {  // Mah, duck-typing is not the best design here. Should separate long press marker from selected search result marker to get rid of this.
                // wrap SwiftUI View in UIView for now, as long as there's no native SwiftUI support
                // => [IOTSDK-5815]
                let ib = InfoBubbleView(titleText: titleText ?? "n/a", addressText: addressText ?? "n/a", telText: telText ?? "n/a")
                let infoBubbleView = UIHostingController(rootView: ib).view

                if infoBubbleView != nil {
                    infoBubbleView?.sizeToFit() // kills transparancy but needed for anchor point
                    infoBubbleView?.backgroundColor = UIColor.clear // set transparancy explicitely now again
                    infoBubbleView?.layer.anchorPoint = CGPoint(x: 0.5, y: 1.0) // set anchorpoint bottom center to peak
                    // Unsure if there's a better way to solve the transparancy issue after swiftui view to uiview conversion

                    self.infoBubble = self.mapView?.pinView(infoBubbleView!, to: topmostMapMarker!.coordinates)
                }
            }

            // place selected marker like long press POI
            self.activeLongPressMarker = MapMarker( at: topmostMapMarker!.coordinates,
                                                  image: MapImage( pixelData: #imageLiteral(resourceName: "Default POI Marker")
                                                                    .resizeImageWith(newSize: CGSize(width: 50,height: 50))
                                                                    .pngData()!, imageFormat: ImageFormat.png),
                                                  anchor: Anchor2D(horizontal: 0.5,vertical: 1.0) )
            self.activeLongPressMarker?.metadata = topmostMapMarker!.metadata // copy metadata too
            self.mapView?.mapScene.addMapMarker( activeLongPressMarker! )

            calculateAndAddRoutesToMap(to: topmostMapMarker!.coordinates)

        } else if pickedMapItems?.polylines.first != nil {  // check if polyline was clicked
            print("Polyline clicked")

            let clickedPolyline = pickedMapItems?.polylines.first
            if clickedPolyline == self.alternativeRoutePolyline {
                swapActiveAndAlternativeRoute()
            }
        } else if pickedMapItems?.polygons.first != nil {       // check if polygon was clicked
            print("Polygon clicked. No action planned here.")
        } else if pickedMapItems?.clusteredMarkers.first != nil { // check if map marker cluster was clicked
               print("Marker cluster clicked. No action planned here.")
        } else {
            print("No supported map marker clicked")
        }

    }


    // @implement LongPressDelegate
    func onLongPress(state: heresdk.GestureState, origin: heresdk.Point2D) {

        guard state == .begin && mapView != nil else { return }

        if self.visualNavigator != nil {
            print("Please stop guidance first...")
            self.alertMessage = "Please stop guidance first before long press"
            self.isAlertShown = true
            return
        }

        print("Long pressed on map \(String(describing: mapView?.viewToGeoCoordinates(viewCoordinates: origin)))")

        // put a marker on long press position
        if self.activeLongPressMarker == nil && self.currentSearchResultMarker.count <= 0 { // ...if not existing
            self.activeLongPressMarker = MapMarker( at: (mapView?.viewToGeoCoordinates(viewCoordinates: origin))!,
                                                  image: MapImage( pixelData: #imageLiteral(resourceName: "Default POI Marker")
                                                                    .resizeImageWith(newSize: CGSize(width: 50,height: 50))
                                                                    .pngData()!, imageFormat: ImageFormat.png),
                                                  anchor: Anchor2D(horizontal: 0.5,vertical: 1.0) )
            self.mapView?.mapScene.addMapMarker( activeLongPressMarker! )
        } else {  // if exists, handle it like CLEAR
            clearMap()
            return
        }

        calculateAndAddRoutesToMap(to: (mapView?.viewToGeoCoordinates(viewCoordinates: origin))!)
    }


    // clear map, remove all mapmarkers
    func clearMap() -> Void {

        if self.activeLongPressMarker != nil {
            self.mapView?.mapScene.removeMapMarker(self.activeLongPressMarker!)
            self.activeLongPressMarker = nil
        }

        if(self.activeRoutePolyline != nil){
            self.mapView?.mapScene.removeMapPolyline( self.activeRoutePolyline! )
            self.activeRoutePolyline = nil
            self.activeRoute = nil
            self.isRouteAvailable = false
            self.currentDestination = nil
        }

        if(self.alternativeRoutePolyline != nil){
            self.mapView?.mapScene.removeMapPolyline( self.alternativeRoutePolyline! )
            self.alternativeRoutePolyline = nil
            self.alternativeRoute = nil
        }

        if self.infoBubble != nil {
            self.infoBubble?.unpin()
            self.infoBubble = nil
        }

        if self.currentSearchResultMarker.count > 0 {

            // remove results from lists
            self.currentSearchResultMarker.removeAll()
            self.searchResultCluster.removeAllMapMarkers()

            // remove cluster from map
            self.mapView?.mapScene.removeMapMarkerCluster(self.searchResultCluster)
        }

    }


    private func addRoutesToMap( routes: [Route]?, zoomToRoute: Bool = true ) -> Void {

        guard routes != nil && routes!.count > 0 else { return }

        // remove old routes if existing
        if self.activeRoutePolyline != nil {
            self.mapView?.mapScene.removeMapPolyline( self.activeRoutePolyline! )
            self.activeRoutePolyline = nil
        }

        if self.alternativeRoutePolyline != nil {
            self.mapView?.mapScene.removeMapPolyline( self.alternativeRoutePolyline! )
             self.alternativeRoutePolyline = nil
        }

        // add new routes
            self.activeRoutePolyline = MapPolyline( geometry: routes![0].geometry,
                                                      widthInPixels: 25.0,
                                                      color: MapController.ACTIVE_COLOR,
                                                      outlineWidthInPixels: 5.0,
                                                      outlineColor: UIColor.white )

            self.activeRoutePolyline?.drawOrder = 1  // make sure main route is on top

            self.mapView?.mapScene.addMapPolyline( self.activeRoutePolyline! )
            self.activeRoute = routes?[0]

            /*
            // DEBUG print
            print("###Section maneuvers: \(routes!.first!.sections.first?.maneuvers.count)")
            print("###Sections poyline vertices: \(routes!.first!.sections.first?.polyline.count)")
            print("###Sections traffic infos: \(routes!.first!.sections.first?.trafficSpeeds.count)")

            // draw traffic on route
            for s in routes!.first!.sections {
                for l in s.trafficSpeeds {
                    var current = s.polyline[Int(l.offset)]
                    var next = s.polyline[Int(l.offset+1)]
                    var jamFactor = Int((l.jamFactor?.rounded())!)
                    if jamFactor > 1 {
                        var trafficPoly = MapPolyline( geometry: try GeoPolyline( vertices: [current, next] ),
                                                       widthInPixels: 15.0,
                                                       color: MapController.JAMCOLORS[jamFactor] )
                        print("Factor: \(jamFactor)")

                        trafficPoly.drawOrder = 2  // make sure traffic info is on very top

                        self.trafficPolylines.append( trafficPoly )

                        self.mapView?.mapScene.addMapPolyline( trafficPoly )
                    }
                }
            }
            */

            // handle also one alternative route if available
            if routes!.count > 1 {
                self.alternativeRoutePolyline = MapPolyline( geometry: routes![1].geometry,
                                                          widthInPixels: 25.0,
                                                          color: MapController.ALTERNATIVE_COLOR,
                                                          outlineWidthInPixels: 5.0,
                                                          outlineColor: UIColor.white )

                // show alternative line as dashed line
                self.alternativeRoutePolyline?.dashPattern = DashPattern(gapLength: 10, dashLength: 30)
                self.alternativeRoutePolyline?.drawOrder = 0

                self.mapView?.mapScene.addMapPolyline( self.alternativeRoutePolyline! )
                self.alternativeRoute = routes?[1]
            }

            self.isRouteAvailable = true

            // visual candy: zoom to route if bbox available
            if zoomToRoute {
                let width = Double(UIScreen.main.bounds.width * UIScreen.main.nativeScale)
                let height = Double(UIScreen.main.bounds.height * UIScreen.main.nativeScale)
                let paddingRect = Rectangle2D( origin: Point2D(x: width*0.1, y: height*0.1), size: Size2D(width: width*0.8, height: height*0.8) ) // add some extra padding to the bounding box
                self.mapView?.camera.lookAt(area: self.activeRoute!.boundingBox, orientation: GeoOrientationUpdate(bearing: 0.0, tilt: 0.0), viewRectangle: paddingRect)
            }

    }


    private func swapActiveAndAlternativeRoute() -> Void {
        let tr = self.activeRoute
        let tp = self.activeRoutePolyline

        self.activeRoute = self.alternativeRoute
        self.activeRoutePolyline = self.alternativeRoutePolyline
        self.activeRoutePolyline?.lineColor = MapController.ACTIVE_COLOR
        self.alternativeRoutePolyline?.dashPattern = nil
        self.activeRoutePolyline?.drawOrder = 1

        self.alternativeRoute = tr
        self.alternativeRoutePolyline = tp
        self.alternativeRoutePolyline?.lineColor = MapController.ALTERNATIVE_COLOR
        self.alternativeRoutePolyline?.dashPattern = DashPattern(gapLength: 5, dashLength: 20)
        self.alternativeRoutePolyline?.drawOrder = 0
    }


    private func calculateRoute(to: GeoCoordinates, routeCompletion: @escaping heresdk.CalculateRouteCompletionHandler)
    {
            guard self.hLocService != nil && self.hLocService?.engine != nil && self.hLocService?.engine?.isStarted  == true && self.hLocService?.engine?.lastKnownLocation != nil else {
                print("Warning: No current position found or service not running.")
                self.alertMessage = "No current position found to calculate route."
                self.isAlertShown = true
                return
            }

            do {
               if self.mapSettings.useOfflineRouting == false {
                        self.routingEngine = try RoutingEngine()
                        print("Using Online routing service...")
               } else {
                        self.routingEngine = try OfflineRoutingEngine()
                        print("Using Offline routing engine...")
               }
            } catch let error {
                print("Failed to initialize router: \(error)")
                self.alertMessage = "ERROR: Could not create router: \(error)"
                self.isAlertShown = true
                return
            }

            // show waiting dialog
            self.waitingDialogText = "Calculating route..."
            self.isWaitingDialogShown = true

            print("Calculate route from \(String(describing: self.hLocService!.engine?.lastKnownLocation)) to \(to)")

            // generic route options used for all route types
            var routeOptions = RouteOptions()
            routeOptions.alternatives = 1 // main route + 1 alternative route requested
            routeOptions.optimizationMode = self.mapSettings.routeOptimization

            // generic text options used for all route types
            var textOptions = RouteTextOptions()
            textOptions.language = self.mapSettings.localeDeviceLanguageMapped ?? LanguageCode.enUs
            textOptions.unitSystem = self.mapSettings.unitSystem

            if self.mapSettings.transportMode == TransportMode.car {
                print("Transport mode is car")
                var options = CarOptions()
                    options.routeOptions = routeOptions
                    options.textOptions = textOptions
                    options.avoidanceOptions.roadFeatures = [RoadFeatures.dirtRoad] // avoid dirty roads

                    self.routingEngine?.calculateRoute( with: [ Waypoint(coordinates: (hLocService!.engine?.lastKnownLocation!.coordinates)!), Waypoint(coordinates: to) ],
                                                        carOptions: options, completion: routeCompletion )  // note: I just forward the completion handler to calling method

            } else if self.mapSettings.transportMode == TransportMode.pedestrian {
                print("Transport mode is pedestrian")
                var options = PedestrianOptions()
                    options.routeOptions = routeOptions
                    options.textOptions = textOptions
                    options.walkSpeedInMetersPerSecond = 1.4  // meter per seconds - 1.4 is aroundd 5 km/h

                    self.routingEngine?.calculateRoute( with: [ Waypoint(coordinates:(hLocService!.engine?.lastKnownLocation!.coordinates)!), Waypoint(coordinates: to) ],
                                                                pedestrianOptions: options, completion: routeCompletion )  // note: I just forward the completion handler to calling method
                
            } else if self.mapSettings.transportMode == TransportMode.truck {
                print("Transport mode is truck")
                var truckSpec = TruckSpecifications()  // using a Mercedes Benz Atego 1218 12T example spec
                    truckSpec.grossWeightInKilograms = 12000
                    truckSpec.heightInCentimeters = 3860
                    truckSpec.widthInCentimeters = 2550
                    truckSpec.lengthInCentimeters = 6800

                var options = TruckOptions()
                    options.routeOptions = routeOptions
                    options.textOptions = textOptions
                    options.avoidanceOptions.roadFeatures = [RoadFeatures.dirtRoad] // avoid dirty roads
                    options.truckSpecifications = truckSpec

                    self.routingEngine?.calculateRoute( with: [ Waypoint(coordinates:(hLocService!.engine?.lastKnownLocation!.coordinates)!), Waypoint(coordinates: to) ],
                                                                truckOptions: options, completion: routeCompletion )  // note: I just forward the completion handler to calling method

            } else if self.mapSettings.transportMode == TransportMode.scooter {
                print("Transport mode is scooter")
                var options = ScooterOptions()
                    options.routeOptions = routeOptions
                    options.textOptions = textOptions
                    options.avoidanceOptions.roadFeatures = [RoadFeatures.dirtRoad] // avoid dirty roads

                    self.routingEngine?.calculateRoute( with: [ Waypoint(coordinates: (hLocService!.engine?.lastKnownLocation!.coordinates)!), Waypoint(coordinates: to) ],
                                                        scooterOptions: options, completion: routeCompletion )  // note: I just forward the completion handler to calling method
            } else if self.mapSettings.transportMode == TransportMode.bicycle {
                print("Transport mode is bicycle")
                var options = BicycleOptions()
                    options.routeOptions = routeOptions
                    options.textOptions = textOptions

                    self.routingEngine?.calculateRoute( with: [ Waypoint(coordinates: (hLocService!.engine?.lastKnownLocation!.coordinates)!), Waypoint(coordinates: to) ],
                                                        bicycleOptions: options, completion: routeCompletion )  // note: I just forward the completion handler to calling method
            }


    }


    private func calculateAndAddRoutesToMap(to: GeoCoordinates) -> Void {

        self.calculateRoute(to: to) { (error: RoutingError?, routes: [Route]?) in

            self.isWaitingDialogShown = false  // always cancel progress dialog
            self.currentDestination = to

            guard error == nil else {
                print("Error: Could not calculate route \(String(describing: error?.name))")
                self.alertMessage = "Could not calculate route \(String(describing: error?.name))"
                self.isAlertShown = true
                return
            }

            // print notice information for all routes
            // notices are information about violation of route calculations
            // These information might be driving safety relevant and should be shown!
            if routes != nil && routes!.count > 0 {
                var noticemessage = ""
                for routeIdx in 0..<routes!.count {
                    for secIdx in 0..<routes![routeIdx].sections.count {
                        for note in routes![routeIdx].sections[secIdx].sectionNotices {
                            print("Routecalculation violated with notice: \(note.code.name) in route #\(routeIdx) and section # \(secIdx)")
                            noticemessage += "=> \(note.code.name) in route #\(routeIdx) and section # \(secIdx)\n"
                        }
                    }
                }

                if( noticemessage != ""){
                    self.alertMessage = "Violations notices for calculated route:\n"+noticemessage
                    self.isAlertShown = true
                }

            }

            print("Route calculated... Got : \(String(describing: routes?.count)) routes from system")

            self.addRoutesToMap(routes: routes)
        }
    }


    func reloadMapStyleSettings() -> Void {
        print("Reloading map style settings")

        // load the map scheme
        //self.mapView?.mapScene.loadScene(fromFile: Utils.getSchemeFileName(mapScheme: self.mapSettings.mapScheme), completion: { (loadSceneError) in
        self.mapView?.mapScene.loadScene(mapScheme: self.mapSettings.mapScheme, completion: { (loadSceneError) in

            if loadSceneError != nil {
                print("Map scene loaded with error: \(String(describing: loadSceneError)) !")
            }

            self.loadMapFeatures()
        })
    }
    
    
    private func loadMapFeatures() -> Void {
        // show traffic on map
        if self.mapSettings.trafficFlowVisible {
            self.mapView?.mapScene.enableFeatures( [MapFeatures.trafficFlow : MapFeatureModes.trafficFlowWithFreeFlow] )
        } else {
            self.mapView?.mapScene.disableFeatures( [MapFeatures.trafficFlow] )
        }

        if self.mapSettings.trafficIncidentsVisible {
            self.mapView?.mapScene.enableFeatures([MapFeatures.trafficIncidents : MapFeatureModes.trafficIncidentsAll])
        } else {
            self.mapView?.mapScene.disableFeatures([MapFeatures.trafficIncidents])
        }

        // extruded buildings
        if self.mapSettings.buildingFootprintsVisible {
            self.mapView?.mapScene.enableFeatures([MapFeatures.buildingFootprints : MapFeatureModes.buildingFootprintsAll])
        } else {
            self.mapView?.mapScene.disableFeatures([MapFeatures.buildingFootprints])
        }
        
        if self.mapSettings.extrudedBuildingsVisible {
            self.mapView?.mapScene.enableFeatures([MapFeatures.extrudedBuildings : MapFeatureModes.extrudedBuildingsAll])
        } else {
            self.mapView?.mapScene.disableFeatures([MapFeatures.extrudedBuildings])
        }

        // truck attributes
        if self.mapSettings.vehicleRestrictionsVisible {
            self.mapView?.mapScene.enableFeatures([MapFeatures.vehicleRestrictions : MapFeatureModes.vehicleRestrictionsActiveAndInactive])
        } else {
            self.mapView?.mapScene.disableFeatures([MapFeatures.vehicleRestrictions])
        }

        // environmental- and congestion zones
        if self.mapSettings.congestionZonesVisible {
            self.mapView?.mapScene.enableFeatures([ MapFeatures.congestionZones : MapFeatureModes.congestionZonesAll,
                                                    MapFeatures.environmentalZones : MapFeatureModes.environmentalZonesAll ])
        } else {
            self.mapView?.mapScene.disableFeatures([ MapFeatures.congestionZones,
                                                     MapFeatures.environmentalZones ])
        }
        
        // 3D landmarks
        if self.mapSettings.landmarksVisible {
            self.mapView?.mapScene.enableFeatures([MapFeatures.landmarks : MapFeatureModes.landmarksTextured])
        } else {
            self.mapView?.mapScene.disableFeatures([MapFeatures.landmarks])
        }

        // Terrain / Hillshading
        if self.mapSettings.hillshadingVisible {
            self.mapView?.mapScene.enableFeatures([MapFeatures.terrain : MapFeatureModes.hillshading])
        } else {
            self.mapView?.mapScene.disableFeatures([MapFeatures.terrain])
        }

        // safety cameras - always show for now!
        self.mapView?.mapScene.enableFeatures([MapFeatures.safetyCameras : MapFeatureModes.safetyCamerasAll])
    }

}
