//
//  MapLoaderController.swift
//  iOS Navigate Reference App
//
//  Created by Jonda, Sachin on 25.04.23.
//  Copyright © 2023 HERE Technologies. All rights reserved.
//

import Foundation
import heresdk

enum MapDownloadStatus {
    case None
    case Done
    case Error
    case Cancelled
    case Paused
    case Progress(Int)
}

class DownloadRegionsStatus: ObservableObject, DownloadRegionsStatusListener {

    @Published var status: MapDownloadStatus = .None

    var cancel: () -> () = {}   // set cancelling via handler in downloadregion method
    var pause: () -> () = {}    // set pausing via handler in downloadregion method
    var resume: () -> () = {}   // set resuming via handler in downloadregion method

    // @implements DownloadRegionsStatusListener
    func onDownloadRegionsComplete(error: heresdk.MapLoaderError?, regions: [heresdk.RegionId]?) {
        if error != nil {
            print("Download completed with error status: \(error?.name ?? "n/a")")
            if error == MapLoaderError.operationCancelled {
                status = .Cancelled
            } else {
                status = .Error
            }
        } else if regions != nil {
            print("Map download completed for regionId #\(String(describing: regions?.description))")
            status = .Done
        }else {
            print("Error: Received completed status for map download, but no error and no region are connected.")
        }
    }

    // @implements DownloadRegionsStatusListener
    func onProgress(region: heresdk.RegionId, percentage: Int32) {
        //DEBUG //print("Download progressing for id#\(region.id) with \(percentage)%")
        status = .Progress(Int(percentage))
    }

    // @implements DownloadRegionsStatusListener
    func onPause(error: MapLoaderError?) {
        print("OnPause called...")
        status = .Paused
    }

    // @implements DownloadRegionsStatusListener
    func onResume() {
        print("OnResume called...")
    }

}


class MapLoaderController : NSObject, ObservableObject {

    var mapSettings: MapSettings?
    var mapLoader: MapDownloader?
    var language: LanguageCode?

    @Published var regionList: [Region]?
    @Published var isRegionListLoading = false
    @Published var installedRegions: [InstalledRegion]?

    init(mapSettings: MapSettings) {
        super.init()

        self.mapSettings = mapSettings
        self.language = mapSettings.localeDeviceLanguageMapped ?? LanguageCode.enUs

        guard let sdkNativeEngine = SDKNativeEngine.sharedInstance else {
            print("Error: SDKNativeEngine not initialized.")
            return
        }
        
        MapDownloader.fromEngineAsync(sdkNativeEngine, { mapDownloaderInstance in
            print("MapController: MapDownloader instance created...")
            
            self.mapLoader = mapDownloaderInstance
            
            self.checkInstallationStatus()
        })

        print("MapLoaderController created with language: \(self.language?.name ?? "n/a")")
    }

    // @implements DownloadRegionsStatusListener
    // gets called when list download is finished, either successful or with an error
    private lazy var downloadRegionListCallback: (_ error: MapLoaderError?, _ result: [Region]?) -> Void = { error, regionList in
        self.isRegionListLoading = false

        guard regionList != nil && regionList!.count > 0 else {
            print("Warning: Got regionList completed callback, but no regions were attached! Reason could be: \(String(describing: error?.name))")
            return
        }

        print("Region download returned: \(regionList!.count) top-level entries with children.")
        self.regionList = regionList
     }

    // @implements DeleteRegionsCompletionHandler
    private lazy var deleteRegionCallback: (MapLoaderError?, [RegionId]?) -> Void = { error, regionList in
        guard error == nil else {
            print("Warning: Deletion of downloaded mapdata failed. Reason: \(String(describing: error?.description))")
            return
        }

        print("Deleted region #\(String(describing: regionList?.description))")
        self.getInstalledRegions() // update installed regions list
    }

    func downloadRegion(region: Region, downloadRegionsStatus: DownloadRegionsStatus) -> Void {
        print("Start downloading map data for region \(region.name) with size \(region.sizeOnNetworkInBytes) / \(region.sizeOnDiskInBytes)")

        guard mapLoader != nil else {
            print("Warning: MapDownloader not ready!")
            return
        }

        let downloaderTask: MapDownloaderTask = (self.mapLoader?.downloadRegions(regions: [region.regionId], statusListener: downloadRegionsStatus))!

        downloadRegionsStatus.cancel = {
            print("Cancelling ongoing download for map region: \(region.name).")
            downloaderTask.cancel()
        }
        downloadRegionsStatus.pause = {
            print("Pausing ongoing download for map region: \(region.name).")
            downloaderTask.pause()
        }
        downloadRegionsStatus.resume = {
            print("Resuming paused download for map region: \(region.name).")
            downloaderTask.resume()
        }

    }

    
    func downloadRegionList() -> Void {
        print("Starting downloading region list...")
        
        guard mapLoader != nil else {
            print("Warning: MapDownloader not ready!")
            return
        }

        self.isRegionListLoading = true

        _ = self.mapLoader?.getDownloadableRegions(languageCode: self.language!, completion: self.downloadRegionListCallback);
    }

    
    func deleteDownloadedMap(regionId: RegionId) {
        print("Delete downloaded map #\(regionId) from disk.")
        
        guard mapLoader != nil else {
            print("Warning: MapDownloader not ready!")
            return
        }

        self.mapLoader?.deleteRegions(regions: [regionId], completion: self.deleteRegionCallback)
    }

    
    func getInstalledRegions() -> Void {
        print("Retrieving already installed map regions from disk.")

        guard mapLoader != nil else {
            print("Warning: MapDownloader not ready!")
            return
        }

        do {
            try self.installedRegions = self.mapLoader?.getInstalledRegions()
        } catch let error {
            print("Error: Could not load installed mapdata regions. Reasons: \(error)")
        }

        print("Installed regions on disk are: \(String(describing: self.installedRegions?.description))")
    }

    
    private func checkInstallationStatus() {
        print("Check persisted map consistency and sanity...")
        
        guard mapLoader != nil else {
            print("Warning: MapDownloader not ready!")
            return
        }

        let persistentMapStatus = self.mapLoader?.getInitialPersistentMapStatus()
            if persistentMapStatus != PersistentMapStatus.ok {
                // Something went wrong after the app was closed the last time. It seems the offline map data is
                // corrupted. This can eventually happen, when an ongoing map download was interrupted due to a crash.
                print("PersistentMapStatus: The persistent map data seems to be corrupted. Trying to repair.")

                // Let's try to repair.
                self.mapLoader?.repairPersistentMap { persistentMapRepairError in
                    if persistentMapRepairError != nil {
                        print("WARNING: RepairPersistentMap: Repair operation failed: \(String(describing: persistentMapRepairError))")
                        return
                    }

                    print("RepairPersistentMap: Repair operation completed successfully!")
                }
            } else {
                print("PersistentMapStatus: The persistent map data seems to be OK.")
            }
    }

}



