//
//  MapLoaderView.swift
//  iOS Navigate Reference App
//
//  Created by Jonda, Sachin on 25.04.23.
//  Copyright © 2023 HERE Technologies. All rights reserved.
//

import SwiftUI
import heresdk

struct RegionRow: View {

    @EnvironmentObject var mapLoaderController: MapLoaderController
    
    @ObservedObject var regionDownloadStatus = DownloadRegionsStatus()

    @State var isExpanded = false
    @State var isInstalled:Bool? = false

    var region: Region!  // region to display for this row
        
    func onRegionExpansionClicked(_ region: Region) {
        print("Expand \(region.name) with id# \(region.regionId.id)")
        self.isExpanded.toggle()
    }

    var body: some View {
        HStack {

            // show region name and expand/collapse buttons
            HStack {
                // prefix with proper icon (expanding / collapsed nor or leaf)
                if region.childRegions != nil && region.childRegions!.count > 0 {
                    if self.isExpanded {
                        Image(systemName: "arrowtriangle.down.fill")
                        .foregroundColor(Color(#colorLiteral(red: 0.6000000238, green: 0.6000000238, blue: 0.6000000238, alpha: 1)))
                        .imageScale(.medium)
                    } else {
                        Image(systemName: "arrowtriangle.right.fill")
                        .foregroundColor(Color(#colorLiteral(red: 1, green: 1, blue: 1, alpha: 1)))
                        .imageScale(.medium)
                    }
                } else {
                    Image(systemName: self.isInstalled==true ?  "checkmark.square.fill" : "circlebadge.fill")
                    .foregroundColor(self.isInstalled==true ? Color(#colorLiteral(red: 0.3411764801, green: 0.6235294342, blue: 0.1686274558, alpha: 1)) : Color(#colorLiteral(red: 0.6000000238, green: 0.6000000238, blue: 0.6000000238, alpha: 1)))
                    .imageScale(.small)
                }

                // print region title name
                Text(region.name)
                    .font(.headline)
                    .foregroundColor(self.isInstalled==true ? Color(#colorLiteral(red: 0.3411764801, green: 0.6235294342, blue: 0.1686274558, alpha: 1)) : Color(#colorLiteral(red: 1, green: 1, blue: 1, alpha: 1)))
                    .lineLimit(2)
            }
            .onTapGesture {
                self.onRegionExpansionClicked(region)                
            }

            Spacer()

            // Show disksize of region
            Text("\(region.sizeOnDiskInBytes/1024/1024)MB")
                .font(.subheadline)
                .lineLimit(1)
                .foregroundColor(self.isInstalled==true ? Color(#colorLiteral(red: 0.3411764801, green: 0.6235294342, blue: 0.1686274558, alpha: 1)) : Color(#colorLiteral(red: 1, green: 1, blue: 1, alpha: 1)))

            // status updates (progress, delete, download buttons)
            switch regionDownloadStatus.status {

                    case .Done:
                        HStack{
                                Image(systemName: "xmark.square.fill")
                                        .foregroundColor(Color(#colorLiteral(red: 0.7450980544, green: 0.1568627506, blue: 0.07450980693, alpha: 1)))
                                        .imageScale(.large)
                        }
                        .onTapGesture { // when clicking the delete button
                                self.mapLoaderController.deleteDownloadedMap(regionId: region.regionId)
                        }

                    case .Error:
                        HStack {
                            Text("Error")
                                .font(.subheadline)
                                .lineLimit(1)
                                .foregroundColor(Color(#colorLiteral(red: 0.7450980544, green: 0.1568627506, blue: 0.07450980693, alpha: 1)))

                            Image(systemName: "repeat.circle.fill")
                                .foregroundColor(Color(#colorLiteral(red: 0.9529411793, green: 0.6862745285, blue: 0.1333333403, alpha: 1)))
                                .imageScale(.large)
                        }
                        .onTapGesture {
                            self.mapLoaderController.downloadRegion(region: region, downloadRegionsStatus: regionDownloadStatus)
                        }

                    case .Progress(let p):
                        HStack {
                            Text("\(p) %")
                                .font(.subheadline)
                                .lineLimit(1)
                                .foregroundColor(.white)

                            Image(systemName: "pause.rectangle.fill")
                                .foregroundColor(Color(#colorLiteral(red: 0.7450980544, green: 0.1568627506, blue: 0.07450980693, alpha: 1)))
                                .imageScale(.large)
                        }
                        .onTapGesture {
                            regionDownloadStatus.cancel()
                        }

                    default:  // None, Cancelled
                        // if region is installed, show delete button (and not download button)
                        if self.isInstalled == true {
                            HStack{
                                    Image(systemName: "xmark.square.fill")
                                            .foregroundColor(Color(#colorLiteral(red: 0.7450980544, green: 0.1568627506, blue: 0.07450980693, alpha: 1)))
                                            .imageScale(.large)
                            }
                            .onTapGesture { // when clicking the delete button
                                    self.mapLoaderController.deleteDownloadedMap(regionId: region.regionId)
                            }
                        } else {
                            HStack {
                                // show download button
                                Image(systemName: "arrow.down.square.fill")
                                    .foregroundColor(Color(#colorLiteral(red: 1, green: 1, blue: 1, alpha: 1)))
                                    .imageScale(.large)
                            }
                            .onTapGesture {
                                self.mapLoaderController.downloadRegion(region: region, downloadRegionsStatus: regionDownloadStatus)
                            }

                        }
            }

        }
        .onAppear {
            // for performance reasons, do the map query only once for each ui update
            self.isInstalled = self.mapLoaderController.installedRegions?.contains(where: { $0.regionId.id == region.regionId.id })
        }

        if region.childRegions != nil && region.childRegions!.count > 0 && self.isExpanded {
            ForEach(region.childRegions!, id: \.self) { childRegion in
                RegionRow(region: childRegion)
                    .padding(.leading)
            }
        }
    }
}


struct MapLoaderView: View {

    @EnvironmentObject var mapLoaderController: MapLoaderController

    var body: some View {
        GeometryReader { geometry in

            VStack(spacing: 0) {

                HStack {
                    Text("Download maps")
                        .font(.title)
                        .foregroundColor(.white)
                        .lineLimit(2)

                        Spacer()

                    Image(systemName: "xmark.circle")
                        .foregroundColor(Color(#colorLiteral(red: 1, green: 1, blue: 1, alpha: 1)))
                        .imageScale(.large)
                }
                .padding(.all)
                .onTapGesture {
                    print("Closing MapLoader dialog")
                    self.mapLoaderController.mapSettings?.showingMapDownloader = false
                }


                if mapLoaderController.isRegionListLoading == true {
                    Spacer()
                    HStack {
                        Spacer()
                            WaitingDialog(titleText: "Loading...")
                                .frame(width: geometry.size.width/1.5, height: geometry.size.height/3)
                        Spacer()
                    }
                    Spacer()
                } else if mapLoaderController.regionList != nil && mapLoaderController.regionList!.count > 0 {
                    ScrollView {
                        ForEach(mapLoaderController.regionList!, id: \.self) { region in
                            RegionRow(region: region)
                                .padding(.bottom, 5)
                                .padding(.horizontal)
                        }
                    }
                }
            }
            .padding(.bottom, geometry.safeAreaInsets.bottom )
            .padding(.top, geometry.safeAreaInsets.top )
            .background(Color(#colorLiteral(red: 0.2549019754, green: 0.2745098174, blue: 0.3019607961, alpha: 1)))
            .edgesIgnoringSafeArea(.all)
            //.frame(width: UIScreen.main.bounds.width, height: UIScreen.main.bounds.height)  // full screen
            .frame(width: geometry.size.width, height: geometry.size.height)  // full screen
            .onAppear {
                self.mapLoaderController.downloadRegionList()
                self.mapLoaderController.getInstalledRegions()
            }

        }
    }
}


struct MapLoaderView_Previews: PreviewProvider {
    static var previews: some View {
        MapLoaderView()
        .environmentObject( MapLoaderController( mapSettings: MapSettings() ) )
    }
}
