//
//  MapSettings.swift
//  iOS Navigate Reference App
//
//  Created by Jonda, Sachin on 12.06.22.
//  Copyright © 2023 HERE Technologies. All rights reserved.
//

import Foundation
import heresdk
import SwiftUI

class MapSettings : NSObject, ObservableObject {

    /* SET IN SIDEMENU DRAWER */
    // selected transportation mode for route calculation and guidance
    @Published var transportMode: TransportMode = TransportMode.car

    // will guidance be simulated with route, or real positioning used
    @Published var simulationMode: Bool = false

    // will guidance session been recorded to GPX trace
    @Published var recordingActive: Bool = false

    // will search/geocoding be offline instead of online ?
    @Published var useOfflineSearch: Bool = false

    // will routing be offline instead of online ?
    @Published var useOfflineRouting: Bool = false

    // is SDK in offline only mode
    @Published var offlineMode: Bool = false

    /* SET IN MAPSCHEME SELECTOR */
    @Published var mapScheme: MapScheme = MapScheme.normalDay
    
    // current visibility of traffic flow and incidents
    @Published var trafficFlowVisible: Bool = true
    @Published var trafficIncidentsVisible: Bool = false

    // extruded buildings
    @Published var extrudedBuildingsVisible: Bool = true
    @Published var buildingFootprintsVisible: Bool = true

    // truck attributes
    @Published var vehicleRestrictionsVisible: Bool = false

    // congestion and environmental zones attributes
    @Published var congestionZonesVisible: Bool = false

    // 3D landmarks
    @Published var landmarksVisible: Bool = false

    // Hillshading / Terrain
    @Published var hillshadingVisible: Bool = true

    /* UI AND FLOW CONTROL */
    @Published var showingSchemeSwitcher = false
    @Published var showingMapDownloader = false
    @Published var showingMapUpdater = false

    /* NOT CONFIGURABLE AT THEN MOMENT */

    // using metric or imperial units
    @Published var unitSystem: UnitSystem = UnitSystem.metric

    // route is optimized for speed or distance
    @Published var routeOptimization: OptimizationMode = OptimizationMode.fastest


    // detect locales
    let localeDeviceRegion = Locale.current.regionCode?.description   // region set on device
    let localeAppLanguage = Locale.current.languageCode  // app language
    let localeDeviceLanguage = Locale.preferredLanguages[0]  // platform device locale, unmapped (used for TTS only in the app)
    let localeDeviceLanguageMapped = Utils.getDeviceLanguageCode()  // device locale mapped to HERE SDK language codes

    // version information
    let appVersion = Bundle.main.infoDictionary?["CFBundleShortVersionString"] as? String
    let appBuildNumber = Bundle.main.infoDictionary?["CFBundleVersion"] as? String


    override init() {
        super.init()
        self.loadSettings()
    }

    func clearCache() -> Void {
        print("Clearing SDK cache (streamed/temporary cache only)...")
        let mapCache = SDKCache.fromEngine(SDKNativeEngine.sharedInstance!);
        mapCache.clearCache { error in
            print("Cleared streamed cache from disk.")
        }
    }

    
    private func loadSettings() {

        // read settings in user defaults
        self.transportMode = TransportMode(rawValue: UInt32(UserDefaults.standard.integer(forKey: "transportMode")) ) ?? TransportMode.car
        
        self.simulationMode = UserDefaults.standard.bool(forKey: "simulationMode")
        self.recordingActive = UserDefaults.standard.bool(forKey: "recordingActive")

        self.useOfflineRouting = UserDefaults.standard.bool(forKey: "useOfflineRouting")
        self.useOfflineSearch = UserDefaults.standard.bool(forKey: "useOfflineSearch")

        self.offlineMode = UserDefaults.standard.bool(forKey: "offlineMode")
        
        
        self.mapScheme = MapScheme( rawValue: UInt32(UserDefaults.standard.integer(forKey: "mapScheme")) ) ?? MapScheme.normalDay

        if UserDefaults.standard.object(forKey: "trafficFlowVisible") != nil {  // only set value if already existing in settings file
            self.trafficFlowVisible =  UserDefaults.standard.bool( forKey: "trafficFlowVisible" )
        }
        
        self.trafficIncidentsVisible = UserDefaults.standard.bool( forKey: "trafficIncidentsVisible" )

        if UserDefaults.standard.object(forKey: "extrudedBuildingVisible") != nil {  // only set value if already existing in settings file
            self.extrudedBuildingsVisible =  UserDefaults.standard.bool( forKey: "extrudedBuildingVisible" )
        }
        
        if UserDefaults.standard.object(forKey: "buildingFootprintsVisible") != nil {  // only set value if already existing in settings file
            self.buildingFootprintsVisible = UserDefaults.standard.bool( forKey: "buildingFootprintsVisible" )
        }
            
        self.vehicleRestrictionsVisible = UserDefaults.standard.bool( forKey: "vehicleRestrictionsVisible" )
 
        self.congestionZonesVisible = UserDefaults.standard.bool( forKey: "congestionZonesVisible" )
        
        self.landmarksVisible = UserDefaults.standard.bool( forKey: "landmarksVisible" )

        if UserDefaults.standard.object(forKey: "hillshadingVisible") != nil {  // only set value if already existing in settings file
            self.hillshadingVisible = UserDefaults.standard.bool( forKey: "hillshadingVisible" )
        }

        print("Setting loaded from disk...")
    }


    public func saveSetting() {

        // write settings in user defaults
        UserDefaults.standard.set(transportMode.rawValue, forKey: "transportMode")

        UserDefaults.standard.set(simulationMode, forKey: "simulationMode")
        UserDefaults.standard.set(recordingActive, forKey: "recordingActive")

        UserDefaults.standard.set(useOfflineRouting, forKey: "useOfflineRouting")
        UserDefaults.standard.set(useOfflineSearch, forKey: "useOfflineSearch")

        UserDefaults.standard.set(offlineMode, forKey: "offlineMode")
        
        UserDefaults.standard.set(mapScheme.rawValue, forKey: "mapScheme")

        UserDefaults.standard.set(trafficFlowVisible, forKey: "trafficFlowVisible")
        UserDefaults.standard.set(trafficIncidentsVisible, forKey: "trafficIncidentsVisible")

        UserDefaults.standard.set(extrudedBuildingsVisible, forKey: "extrudedBuildingVisible")
        UserDefaults.standard.set(buildingFootprintsVisible, forKey: "buildingFootprintsVisible")

        UserDefaults.standard.set(vehicleRestrictionsVisible, forKey: "vehicleRestrictionsVisible")

        UserDefaults.standard.set(congestionZonesVisible, forKey: "congestionZonesVisible")

        UserDefaults.standard.set(landmarksVisible, forKey: "landmarksVisible")
        
        UserDefaults.standard.set(hillshadingVisible, forKey: "hillshadingVisible")
        
        print("Settings saved to disk...")
    }

}
