//
//  MapUpdaterController.swift
//  iOS Navigate Reference App
//
//  Created by Jonda, Sachin on 25.04.23.
//  Copyright © 2023 HERE Technologies. All rights reserved.
//

import Foundation
import heresdk


class MapUpdateStatus: ObservableObject, MapUpdateProgressListener {

    @Published var status: MapDownloadStatus = .None

    var cancel: () -> () = {}   // set cancelling via handler in downloadregion method
    var pause: () -> () = {}    // set pausing via handler in downloadregion method
    var resume: () -> () = {}   // set resuming via handler in downloadregion method

    // @implements onComplete
    func onComplete(error: MapLoaderError?) {
        if error != nil {
            print("Update completed with error status: \(error?.name ?? "n/a")")
            if error == MapLoaderError.operationCancelled {
                status = .Cancelled
            } else {
                status = .Error
            }
        } else {
            print("Map update completed!")
            status = .Done
        }
    }

    // @implements onProgress
    func onProgress(region: heresdk.RegionId, percentage: Int32) {
        //DEBUG
        print("Update progressing for id#\(region.id) with \(percentage)%")
        status = .Progress(Int(percentage))
    }

    // @implements onPause
    func onPause(error: MapLoaderError?) {
        print("OnPause called...")
        status = .Paused
    }

    // @implements DownloadRegionsStatusListener
    func onResume() {
        print("OnResume called...")
    }

}



class MapUpdaterController : NSObject, ObservableObject {

    @Published var isMapUpdateCheckRunning = false
    @Published var updateAvailability: MapUpdateAvailability = .notAvailable

    var mapSettings: MapSettings?
    var mapUpdater: MapUpdater?
    
    init(mapSettings: MapSettings) {
        super.init()
        
        self.mapSettings = mapSettings
        
        guard let sdkNativeEngine = SDKNativeEngine.sharedInstance else {
            print("Error: SDKNativeEngine not initialized.")
            return
        }
        
        MapUpdater.fromEngineAsync( sdkNativeEngine, { mapUpdaterInstance in
            print("MapUpdaterController: MapUpdater instance created...")            
            self.mapUpdater = mapUpdaterInstance
        })

        print("MapUpdateController created...")
    }
    
    
    func checkForMapUpdates() -> Void {
        print("MapUpdater: Checking for new map version...")
        
        guard mapUpdater != nil else {
            print("Warning: MapUpdater not ready!")
            return
        }
        
        self.isMapUpdateCheckRunning = true
        _ = self.mapUpdater?.checkMapUpdate(completion: onMapUpdateCheckCompleted)
    }
        
    
    // Completion handler to get notified whether a map update is available or not.
    private func onMapUpdateCheckCompleted(mapLoaderError: MapLoaderError?, mapUpdateAvailability: MapUpdateAvailability?) {

        self.isMapUpdateCheckRunning = false
        
        guard mapUpdateAvailability != nil else {
            print("MapUpdateCheck Error: UpdateAvailability is nil. Authentication might have failed.")
            return
        }

        self.updateAvailability = mapUpdateAvailability!

        if let error = mapLoaderError {
            print("MapUpdateCheck Error: \(error)")
            return
        }

        do {
            try print("Current map version: \(String(describing: self.mapUpdater?.getCurrentMapVersion().stringRepresentation(separator: ".")))")
        } catch {
            print("Error retrieving map version: error# \(error)")
        }

        if (mapUpdateAvailability == MapUpdateAvailability.available) {
            print("MapUpdateCheck: One or more map updates are available.")
        }else{
            print("MapUpdateCheck: No map update available. Latest versions are already installed.");
        }
    }

    func updateMap(mapUpdateStatus: MapUpdateStatus) -> Void {
        print("Starting map updates...")

        guard mapUpdater != nil else {
            print("Warning: MapUpdater not ready!")
            return
        }

        let updateCheckHandler = self.mapUpdater?.performMapUpdate(completion: mapUpdateStatus)

        mapUpdateStatus.cancel = {
            print("Cancelling ongoing map updates.")
            updateCheckHandler?.cancel()
        }

        mapUpdateStatus.pause = {
            print("Pausing map updates.")
            updateCheckHandler?.pause()
        }

        mapUpdateStatus.resume = {
            print("Resuming paused map updates.")
            updateCheckHandler?.resume()
        }

    }

    
    func getInstalledMapVersion() -> String {
        print("Getting installed and used map version...")

        var version: String = "n/a"

        if self.mapUpdater != nil {
            do {
                version = try self.mapUpdater!.getCurrentMapVersion().stringRepresentation(separator: ",")
            } catch {
                print("Error retrieving map version: error# \(error)")
            }
        }
                
        return version
    }

}
