//
//  SearchBar.swift
//  iOS Navigate Reference App
//
//  Created by Jonda, Sachin on 11.06.22.
//  Copyright © 2023 HERE Technologies. All rights reserved.
//

import SwiftUI

struct TextFieldClearButton: ViewModifier {
    @Binding var text: String
    @ObservedObject var mapController: MapController

    func body(content: Content) -> some View {
        HStack {
            content

            if !text.isEmpty {
                Button(
                    action: {
                        self.text = ""
                        self.mapController.currentSuggestionsList.removeAll()
                    },
                    label: {
                        Image(systemName: "delete.left")
                            .foregroundColor(Color(UIColor.opaqueSeparator))
                    }
                )
            }
        }
    }
}


struct SearchBar: View {

    @ObservedObject var mapController: MapController
    @Binding var menuOpen: Bool
    @State var searchInput = ""
    @State private var inEditMode = false

    var body: some View {

        VStack {
            HStack {
                Button(action: {
                    withAnimation {
                        self.menuOpen.toggle()
                    }
                }) {
                    Image(systemName: "line.horizontal.3")
                        .resizable()
                        .frame(width: 30, height: 20)
                        .foregroundColor(.white)
                        .padding(.leading)
                }

                TextField("Search",
                          text: Binding( get: { return self.searchInput },
                                         set: {  // fire suggestion requests when search input changes
                                                self.searchInput = $0
                                                if self.searchInput.count >= 3 {  // only fire suggest with 3 or more characters input
                                                    self.mapController.suggest(searchTerm: self.searchInput)
                                                } else {
                                                    self.mapController.currentSuggestionsList.removeAll()
                                                }
                                              }),
                          onEditingChanged: {edit in
                                print("Editing in progress: \(edit)")
                                self.inEditMode = edit
                                if edit == false {
                                    // close keyboard
                                    UIApplication.shared.sendAction(#selector(UIResponder.resignFirstResponder), to:nil, from:nil, for:nil)
                                    self.mapController.search(searchTerm: self.searchInput)
                                }
                })
                .modifier(TextFieldClearButton(text: $searchInput, mapController: self.mapController))
                .textFieldStyle(RoundedBorderTextFieldStyle())
                .foregroundColor( Color.black )
                .padding(.all, 10.0)

                Button(action: {
                    // close keyboard
                    UIApplication.shared.sendAction(#selector(UIResponder.resignFirstResponder), to:nil, from:nil, for:nil)
                    self.mapController.search(searchTerm: self.searchInput)
                }) {
                    Image(systemName: "magnifyingglass.circle")
                        .resizable()
                        .frame(width: 30.0, height: 30.0 )
                        .foregroundColor(.white)
                }
                .padding(.trailing, 20.0)
            }
            .frame(height: 70.0)

            // show auto suggest drop down / selection list
            if inEditMode {
                List(self.mapController.currentSuggestionsList, id: \.title) { result in
                    Button(action: {  // takeover suggested result and search for it
                        self.searchInput = result.place?.address.addressText ?? result.title
                        UIApplication.shared.sendAction(#selector(UIResponder.resignFirstResponder), to:nil, from:nil, for:nil)
                    }, label: {
                        Text("\(result.place?.address.addressText ?? result.title)")
                    })
                }
            }
        }
    }
}

struct SearchBar_Previews: PreviewProvider {
    static var previews: some View {
        SearchBar(mapController: MapController(settings: MapSettings()), menuOpen: .constant(true))
    }
}
