import SwiftUI
import UIKit
import heresdk
import CoreLocation
import AVFAudio

class Utils {

    // Returns a human readable string of properly formatted distance.
    // Distance is in meters if < 1000 m, e.g. 123 m
    // Distance is in km if >= 1000m and < 10000, eg 2.3 km
    // Distance is in km if => 10000m, eg 10 km
    public static func formatDistance(meter: Int) -> String {
        if meter < 1000 {
            return String("\(meter) m")
        } else if meter < 10000 {
            return String(format: "%.1f km", Double(meter)/1000.0)
        } else {
            return String(format: "%.0f km", Double(meter)/1000.0)
        }
    }

    // Retuns a human readable string of properly formatted distance.
    // Time in seconds, if < 60s, e.g. 35s
    // Time in minutes is if between 60seconds and 59m , e.g. 34m
    // Time in h and min if >= 60 min, 2:35h
    public static func formatTime(seconds: Int) -> String {
         if seconds < 60 {
                 return String("\(seconds) s")
         } else if seconds < 3600 {
                 return String("\(seconds/60) m")
         } else {
                let (h, m) = (seconds/60).quotientAndRemainder(dividingBy: 60)
                return String("\(h):\(m) h")
        }
    }

    public static let METER_PER_SECOND_TO_KMH = 3.6

    public static func getIconNameforManeuver(action: ManeuverAction) -> String {

        switch action {

        // Departure maneuver, such as "Start at".
        case .depart:
            return "maneuver_depart"

        /// Arrival maneuver, such as "Arrive at".
        case .arrive:
            return "maneuver_arrive"

        /// Left-hand U-turn maneuver, such as "Make a U-turn at".
        case .leftUTurn:
            return "maneuver_left_u_turn"

        /// Sharp left turn maneuver, such as "Make a hard left turn onto".
        case .sharpLeftTurn:
            return "maneuver_sharp_left_turn"

        /// Left turn maneuver, such as "Turn left on".
        case .leftTurn:
            return "maneuver_left_turn"

        /// Slight left turn maneuver, such as "Bear left onto".
        case .slightLeftTurn:
            return "maneuver_slight_left_turn"

        /// Continue maneuver, such as "Continue on".
        case .continueOn:
            return "maneuver_continue_on"

        /// Slight right turn maneuver, such as "Bear right onto".
        case .slightRightTurn:
            return "maneuver_slight_right_turn"

        /// Right turn maneuver, such as "Turn right on".
        case .rightTurn:
            return "maneuver_right_turn"

        /// Sharp right turn maneuver, such as "Make a hard right turn onto".
        case .sharpRightTurn:
            return "maneuver_sharp_right_turn"

        /// Right u-turn maneuver, such as "Make a right U-turn at".
        case .rightUTurn:
            return "maneuver_right_u_turn"

        /// Left exit maneuver, such as "Take the left exit to".
        case .leftExit:
            return "maneuver_left_exit"

        /// Right exit maneuver, such as "Take the right exit to".
        case .rightExit:
            return "maneuver_right_exit"

        /// Left ramp maneuver, such as "Take the left ramp onto".
        case .leftRamp:
            return "maneuver_left_ramp"

        /// Right ramp maneuver, such as "Take the right ramp onto".
        case .rightRamp:
            return "maneuver_right_ramp"

        /// Left fork maneuver, such as "Take the left fork onto".
        case .leftFork:
            return "maneuver_left_fork"

        /// Middle fork maneuver, such as "Take the middle fork onto".
        case .middleFork:
            return "maneuver_middle_fork"

        /// Right fork maneuver, such as "Take the right fork onto".
        case .rightFork:
            return "maneuver_right_fork"

        /// Ferry maneuver, such as "Take the ferry to".
        case .ferry:
            return "maneuver_ferry"

        /// Roundabout maneuver (right-hand traffic), such as "Enter the roundabout".
        case .leftRoundaboutEnter:
            return "maneuver_bear_left"  // no dedicated icon? bear seems fit best.

        /// Roundabout maneuver, such as "Enter the roundabout".
        case .rightRoundaboutEnter:
            return "maneuver_bear_right"  // no dedicated icon? bear seems fit best.

        /// Roundabout maneuver (left-hand traffic), such as "Pass the roundabout".
        case .leftRoundaboutPass:
            return "maneuver_left_roundabout_pass"

        /// Roundabout maneuver, such as "Pass the roundabout".
        case .rightRoundaboutPass:
            return "maneuver_right_roundabout_pass" // no icon?

        /// Roundabout maneuver (left-hand traffic), such as  "Take the first exit of the roundabout onto".
        case .rightRoundaboutExit1:
            return "maneuver_right_roundabout_exit1"

        /// Roundabout maneuver (left-hand traffic), such as "Take the second exit of the roundabout onto".
        case .rightRoundaboutExit2:
            return "maneuver_right_roundabout_exit2"

        /// Roundabout maneuver (left-hand traffic), such as "Take the third exit of the roundabout onto".
        case .rightRoundaboutExit3:
            return "maneuver_right_roundabout_exit3"

        /// Roundabout maneuver (right-hand traffic), such as "Take the fourth exit of the roundabout onto".
        case .rightRoundaboutExit4:
            return "maneuver_right_roundabout_exit4"

        /// Roundabout maneuver (left-hand traffic), such as "Take the fifth exit of the roundabout onto".
        case .rightRoundaboutExit5:
            return "maneuver_right_roundabout_exit5"

        /// Roundabout maneuver (left-hand traffic), such as "Take the sixth exit of the roundabout onto".
        case .rightRoundaboutExit6:
            return "maneuver_right_roundabout_exit6"

        /// Roundabout maneuver (left-hand traffic), such as "Take the 7th exit of the roundabout onto".
        case .rightRoundaboutExit7:
            return "maneuver_right_roundabout_exit7"

        /// Roundabout maneuver (left-hand traffic), such as "Take the 8th exit of the roundabout onto".
        case .rightRoundaboutExit8:
            return "maneuver_right_roundabout_exit8"

        /// Roundabout maneuver (left-hand traffic), such as "Take the 9th exit of the roundabout onto".
        case .rightRoundaboutExit9:
            return "maneuver_right_roundabout_exit9"

        /// Roundabout maneuver (left-hand traffic), such as "Take the 10th exit of the roundabout onto".
        case .rightRoundaboutExit10:
            return "maneuver_right_roundabout_exit10"

        /// Roundabout maneuver (left-hand traffic), such as "Take the 11th exit of the roundabout onto".
        case .rightRoundaboutExit11:
            return "maneuver_right_roundabout_exit11"

        /// Roundabout maneuver (left-hand traffic), such as "Take the 12th exit of the roundabout onto".
        case .rightRoundaboutExit12:
            return "maneuver_right_roundabout_exit12"

        /// Roundabout maneuver (right-hand traffic), such as "Take the first exit of the roundabout onto".
        case .leftRoundaboutExit1:
            return "maneuver_left_roundabout_exit1"

        /// Roundabout maneuver (right-hand traffic), such as "Take the second exit of the roundabout onto".
        case .leftRoundaboutExit2:
            return "maneuver_left_roundabout_exit2"

        /// Roundabout maneuver (right-hand traffic), such as "Take the third exit of the roundabout onto".
        case .leftRoundaboutExit3:
            return "maneuver_left_roundabout_exit3"

        /// Roundabout maneuver (right-hand traffic), such as "Take the fourth exit of the roundabout onto".
        case .leftRoundaboutExit4:
            return "maneuver_left_roundabout_exit4"

        /// Roundabout maneuver (right-hand traffic), such as "Take the fifth exit of the roundabout onto".
        case .leftRoundaboutExit5:
            return "maneuver_left_roundabout_exit5"

        /// Roundabout maneuver (right-hand traffic), such as "Take the sixth exit of the roundabout onto".
        case .leftRoundaboutExit6:
            return "maneuver_left_roundabout_exit6"

        /// Roundabout maneuver (right-hand traffic), such as "Take the 7th exit of the roundabout onto".
        case .leftRoundaboutExit7:
            return "maneuver_left_roundabout_exit7"

        /// Roundabout maneuver (right-hand traffic), such as "Take the 8th exit of the roundabout onto".
        case .leftRoundaboutExit8:
            return "maneuver_left_roundabout_exit8"

        /// Roundabout maneuver (right-hand traffic), such as "Take the 9th exit of the roundabout onto".
        case .leftRoundaboutExit9:
            return "maneuver_left_roundabout_exit9"

        /// Roundabout maneuver (right-hand traffic), such as "Take the 10th exit of the roundabout onto".
        case .leftRoundaboutExit10:
            return "maneuver_left_roundabout_exit10"

        /// Roundabout maneuver (right-hand traffic), such as "Take the 11th exit of the roundabout onto".
        case .leftRoundaboutExit11:
            return "maneuver_left_roundabout_exit11"

        /// Roundabout maneuver (right-hand traffic), such as "Take the 12th exit of the roundabout onto".
        case .leftRoundaboutExit12:
            return "maneuver_left_roundabout_exit12"

        default:
            return "Questionmark"
        }
    }


    // Return SDK's LanguageCode for current device locale, or nil if input language not supported
    public static func getDeviceLanguageCode() -> LanguageCode? {

                let deviceLocale = Locale.preferredLanguages[0]

                switch deviceLocale {

                // Arabic - iOS13+ only
                case "ar-SA", "ar":
                    return LanguageCode.arSa

                // Bulgarian
                case "bg-BG", "bg":
                        return LanguageCode.bgBg

                // Czech
                case "cs-CZ", "cs":
                    return LanguageCode.csCz

                 // Danish
                case "da-DK", "da":
                    return LanguageCode.daDk

                // German
                case "de-DE", "de":
                    return LanguageCode.deDe

                // Greek
                case "el-GR", "el":
                    return LanguageCode.elGr

                // English (United States)
                case "en-US", "en":
                    return LanguageCode.enUs

                // British English
                case "en-GB":
                    return LanguageCode.enGb

                // Spanish
                case "es-ES", "es-AR", "es":
                     return LanguageCode.esEs

                // Mexican Spanish
                case "es-MX":
                    return LanguageCode.esMx

                 // Farsi / Persian
                 case "fa-IR", "fa":
                      return LanguageCode.faIr

                // Finnish
                case "fi-FI", "fi":
                    return LanguageCode.fiFi

                // French
                case "fr-FR", "fr":
                    return LanguageCode.frFr

                // French Canadian
                case "fr-CA":
                    return LanguageCode.frCa

                // Hebrew  - iOS13+ only
                case "he-IL", "he":
                    return LanguageCode.heIl

                // Hindi
                case "hi-IN", "hi":
                    return LanguageCode.hiIn

                // Croatia
                case "hr-HR", "hr":
                    return LanguageCode.hrHr

                // Hungarian
                case "hu-HU", "hu":
                    return LanguageCode.huHu

                // Indonesian
                case "id-ID", "id":
                    return LanguageCode.idId

                // Italian
                case "it-IT", "it":
                    return LanguageCode.itIt

                // Japanese
                case "ja-JP", "ja":
                    return LanguageCode.jaJp

                // Korean
                case "ko-KR", "ko":
                    return LanguageCode.koKr

                // Norwegian / Bokmal
                case "nb-NO", "nb":
                    return LanguageCode.nbNo

                // Dutch
                case "nl-NL", "nl":
                    return LanguageCode.nlNl

                // Polish
                case "pl-PL", "pl":
                    return LanguageCode.plPl

                // Portuguese
                case "pt-PT", "pt":
                    return LanguageCode.ptPt

                // Brasilian Protuguese
                case "pt-BR":
                    return LanguageCode.ptBr

                // Romanian
                case "ro-RO", "ro":
                    return LanguageCode.roRo

                // Russian
                case "ru-RU", "ru":
                    return LanguageCode.ruRu

                // Slovak
                case "sk-SK", "sk":
                    return LanguageCode.skSk

                // Swedish
                case "sv-SE", "sv":
                    return LanguageCode.svSe

                // Turkish
                case "th-TH", "th":
                    return LanguageCode.trTr

                // Thai
                case "tr-TR", "tr":
                    return LanguageCode.thTh

                // Ukrainian
                case "uk-UA", "uk":
                    return LanguageCode.ukUa

                // Chinese Mainland, simplified
                case "zh-CN", "zh-Hans", "zh":
                    return LanguageCode.zhCn

                // Chinese Hong Kong
                case "zh-HK":
                    return LanguageCode.zhHk

                // Chinese Taiwan
                case "zh-TW":
                    return LanguageCode.zhTw

                default:
                    return nil // input language not supported
            }
    }


    /*
    // get scheme filename. use filename if non standard enum map schemes should be used.
    // temporary for japan transition and oslo style
    public static func getSchemeFileName(mapScheme: MapScheme) -> String {

        switch mapScheme {
            case MapScheme.normalNight:
                return "preview.normal.night.json"

            case MapScheme.hybridDay:
                return "preview.hybrid.day.json"

            default:  // normal day and the rest. not for generic use, since i miss some cases not used in my demo app!
                //return "preview.normal.day.json"
                return "japan.ocm.day.json"

        }
    }
    */


}


// =========== EXTENSIONS ===========

extension Location {
    func withoutAltitude() -> Location {
        var newLocWithoutAltitude = Location( coordinates: GeoCoordinates(latitude: coordinates.latitude, longitude: coordinates.longitude, altitude: 0.0) )
        newLocWithoutAltitude.horizontalAccuracyInMeters = horizontalAccuracyInMeters
        newLocWithoutAltitude.verticalAccuracyInMeters = verticalAccuracyInMeters
        newLocWithoutAltitude.bearingAccuracyInDegrees = bearingAccuracyInDegrees
        newLocWithoutAltitude.bearingInDegrees = bearingInDegrees
        newLocWithoutAltitude.speedInMetersPerSecond = speedInMetersPerSecond
        newLocWithoutAltitude.speedAccuracyInMetersPerSecond = speedAccuracyInMetersPerSecond

        return newLocWithoutAltitude
    }
}



// ----- helper extension to scale UIImage on the fly -----
extension UIImage {
    func resizeImageWith(newSize: CGSize) -> UIImage {
        let horizontalRatio = newSize.width / size.width
        let verticalRatio = newSize.height / size.height

        let ratio = max(horizontalRatio, verticalRatio)
        let newSize = CGSize(width: size.width * ratio, height: size.height * ratio)
        UIGraphicsBeginImageContextWithOptions(newSize, false, 0)
        draw(in: CGRect(origin: CGPoint(x: 0, y: 0), size: newSize))
        let newImage = UIGraphicsGetImageFromCurrentImageContext()
        UIGraphicsEndImageContext()
        return newImage!
    }
}

/*
 Expand Apple core location class to return also HERELocation class
 */
extension CLLocation {
    func toHERELocation() -> Location {
        return Location( coordinates: GeoCoordinates(latitude: self.coordinate.latitude, longitude: self.coordinate.longitude, altitude: self.altitude),
                         bearingInDegrees: self.course,
                         speedInMetersPerSecond: self.speed,
                         horizontalAccuracyInMeters: self.horizontalAccuracy,
                         verticalAccuracyInMeters: self.verticalAccuracy)
    }
}

/**
 * Extend enums with method to derive names
 */
extension RawRepresentable where RawValue: Any {
    /**
     * Get Name of enumeration (as written in case clause).
     */
    var name: String {
        get { return String(describing: self) }
    }

    /**
     * Get full name of enumeration
     */
    var description: String {
        get { return String(reflecting: self) }
    }
}


// custom view extension to use partially rounded corners on views:
// https://stackoverflow.com/questions/56760335/round-specific-corners-swiftui
extension View {
    func cornerRadius(_ radius: CGFloat, corners: UIRectCorner) -> some View {
        clipShape( RoundedCorner(radius: radius, corners: corners) )
    }
}

struct RoundedCorner: Shape {

    var radius: CGFloat = .infinity
    var corners: UIRectCorner = .allCorners

    func path(in rect: CGRect) -> Path {
        let path = UIBezierPath(roundedRect: rect, byRoundingCorners: corners, cornerRadii: CGSize(width: radius, height: radius))
        return Path(path.cgPath)
    }
}


// react on change events for segmented picker (only native in ios14 and higher)
// see https://stackoverflow.com/questions/57518852/swiftui-picker-onchange-or-equivalent/60130311#60130311
extension Binding {
    func onChange(_ handler: @escaping (Value) -> Void) -> Binding<Value> {
        return Binding(
            get: { self.wrappedValue },
            set: { selection in
                self.wrappedValue = selection
                handler(selection)
        })
    }
}

